'''
Visual Stimulus codebase implements several classes to display stimulus routines.
Can display frame by frame or compress data for certain stimulus routines and
display by index. Used to manage information between experimental devices and
interact with `StimulusRoutines` to produce visual display and log data. May also
be used to save and export movies of experimental stimulus routines for
presentation.
'''
from psychopy import visual, event
import PIL
import os
import datetime
import tifffile as tf
import numpy as np
import matplotlib.pyplot as plt
import time
from tools import FileTools as ft
from tools.IO import nidaq as iodaq


def analyze_frames(ts_start, ts_end, refresh_rate, check_point=(0.02, 0.033, 0.05, 0.1)):
    """
    Analyze frame durations of time stamp data.

    Computes relevant statistics with respect to the presentation
    of a given stimulus. The statistics are computed in order
    to understand the timing of the frames since the monitor refresh
    rate isn't always an accurate tool for timing.

    Parameters
    ----------
    ts_start : 1d array
        list of time stamps of each frame start (in seconds).
    ts_end: 1d array
        list of time stamps of each frame end (in seconds).
    refresh_rate : float
        the refresh rate of imaging monitor measured (in Hz).
    check_point : tuple, optional

    Returns
    -------
    frame_duration : ndarray
        list containing the length of each time stamp.
    frame_stats : str
        string containing a statistical analysis of the image frames.

    """

    frame_interval = np.diff(ts_start)
    plt.figure()
    plt.hist(frame_interval, bins=np.linspace(0.0, 0.05, num=51))
    refresh_rate = float(refresh_rate)

    num_frames = ts_start.shape[0]
    disp_true = ts_end[-1] - ts_start[0]
    disp_expect = float(num_frames) / refresh_rate
    avg_frame_time = np.mean(frame_interval) * 1000
    sdev_frame_time = np.std(frame_interval) * 1000
    short_frame = min(frame_interval) * 1000
    short_frame_ind = np.where(frame_interval == np.min(frame_interval))[0][0]
    long_frame = max(frame_interval) * 1000
    long_frame_ind = np.where(frame_interval == np.max(frame_interval))[0][0]

    frame_stats = ''
    frame_stats += '\nTotal number of frames      : {}.'.format(num_frames)
    frame_stats += '\nTotal length of display     : {:.5f} second.'.format(disp_true)
    frame_stats += '\nExpected length of display  : {:.5f} second.'.format(disp_expect)
    frame_stats += '\nMean of frame intervals     : {:.2f} ms.'.format(avg_frame_time)
    frame_stats += '\nS.D. of frame intervals     : {:.2f} ms.'.format(sdev_frame_time)
    frame_stats += '\nShortest frame: {:.2f} ms, index: {}.'.format(short_frame, short_frame_ind)
    frame_stats += '\nLongest frame : {:.2f} ms, index: {}.'.format(long_frame, long_frame_ind)

    for i in range(len(check_point)):
        check_number = check_point[i]
        frame_number = len(frame_interval[frame_interval > check_number])
        frame_stats += '\nNumber of frames longer than {:5.3f} second: {}; {:6.2f}%'. \
            format(check_number, frame_number, (float(frame_number) * 100 / num_frames))

    print(frame_stats)

    return frame_interval, frame_stats


class DisplaySequence(object):
    """
    Display the stimulus routine from memory.

    Takes care of high level management of your computer
    hardware with respect to its interactions within a given experiment.
    Stimulus presentation routines are specified and external connection
    to National Instuments hardware devices is provided. Also takes care
    of the logging of relevant experimental data collected and where it
    will be stored on the computer used for the experiment.


    """

    def __init__(self,
                 log_dir,
                 backupdir=None,
                 identifier='000',
                 display_iter=1,
                 mouse_id='Test',
                 user_id='Name',
                 psychopy_mon='testMonitor',
                 is_by_index=False,
                 is_interpolate=False,
                 is_triggered=False,
                 is_save_sequence=False,
                 trigger_event="negative_edge",
                 trigger_NI_dev='Dev1',
                 trigger_NI_port=1,
                 trigger_NI_line=0,
                 is_sync_pulse=True,
                 sync_pulse_NI_dev='Dev1',
                 sync_pulse_NI_port=1,
                 sync_pulse_NI_line=1,
                 display_screen=0,
                 initial_background_color=0.,
                 color_weights=(1., 1., 1.)):
        """
        initialize `DisplaySequence` object

        Parameters
        ----------
        log_dir : str
            system directory path to where log display will be saved.
        backupdir : str, optional
            copy of directory path to save backup, defaults to `None`.
        identifier: str, optional
            identifing string for this particular experiment, this will
            show up in the name of log file when display is done.
        display_iter : int, optional
            defaults to `1`
        mouse_id : str, optional
            label for mouse, defaults to 'Test'.
        user_id : str, optional
            label for person performing experiment, defaults to 'Name'.
        psychopy_mon : str, optional
            label for monitor used for displaying the stimulus, defaults to
            'testMonitor'.
        is_interpolate : bool, optional
            defaults to `False`.
        is_triggered : bool, optional
            if `True`, stimulus will not display until triggered. if `False`,
            stimulus will display automatically. defaults to `False`.
        is_by_index : bool, optional
            determines if stimulus is displayed by index which saves memory
            and should speed up routines. Note that not every stimulus can be
            displayed by index and hence the default value is `False`.
        is_save_sequence : bool, optional
            defaults to False
            if True, the class will save the sequence of images to be displayed
            as a tif file, in the same folder of log file. If self.is_by_index
            is True, only unique frames will be saved. Note, this will save
            the whole sequence even if the display is interrupted in the middle.
        trigger_NI_dev : str, optional
            defaults to 'Dev1'.
        trigger_NI_port : int, optional
            defaults to `1`.
        trigger_NI_line : int, optional
            defaults to `0`.
        is_sync_pulse : bool, optional
            defaults to `True`.
        sync_pulse_NI_dev : str, optional
            defaults to 'Dev1'.
        sync_pulse_NI_port : int, optional
            defaults to 1.
        sync_pulse_NI_line : int, optional
            defaults to 1.
        trigger_event : str
            should be one of "negative_edge", "positive_edge", "high_level",
            or "low_level". defaults to "negative_edge".
        display_screen : int
            determines which monitor to display stimulus on. defaults to `0`.
        initial_background_color : float
            defaults to `0`. should be in the range from -1. (black) to 1. (white)
        color_weights : tuple, optional
            defaults to (1., 1., 1.)
            This should be a tuple with 3 elements. Each element specifies the
            weight of each color channel (R, G, B). The value range of each
            element is in the range [0., 1.]. This is designed in such way that
            if you want to suppress a certain channel i.e. red channel, you can
            change this parameter to (0., 1., 1.)
        """

        self.sequence = None
        self.seq_log = {}
        self.identifier = str(identifier)
        self.psychopy_mon = psychopy_mon
        self.is_interpolate = is_interpolate
        self.is_triggered = is_triggered
        self.is_by_index = is_by_index
        self.is_save_sequence = is_save_sequence
        self.trigger_NI_dev = trigger_NI_dev
        self.trigger_NI_port = trigger_NI_port
        self.trigger_NI_line = trigger_NI_line
        self.trigger_event = trigger_event
        self.is_sync_pulse = is_sync_pulse
        self.sync_pulse_NI_dev = sync_pulse_NI_dev
        self.sync_pulse_NI_port = sync_pulse_NI_port
        self.sync_pulse_NI_line = sync_pulse_NI_line
        self.display_screen = display_screen
        self.initial_background_color = float(initial_background_color)

        if len(color_weights) != 3:
            raise ValueError('input color_weights should be a tuple with 3 elements.')
        for cw in color_weights:
            if cw < -1. or cw > 1.:
                raise ValueError('each element of color_weight should be no less than -1. and no greater than 1.')
        self.color_weights = color_weights

        self.keep_display = None

        if display_iter % 1 == 0:
            self.display_iter = display_iter
        else:
            raise ArithmeticError, "`display_iter` should be a whole number."

        self.log_dir = log_dir
        self.backupdir = backupdir
        self.mouse_id = mouse_id
        self.user_id = user_id
        self.seq_log = None

        self.clear()

    def set_any_array(self, any_array, log_dict=None):
        """
        to display any numpy 3-d array.
        """
        if len(any_array.shape) != 3:
            raise LookupError, "Input numpy array should have dimension of 3!"

        vmax = np.amax(any_array).astype(np.float32)
        vmin = np.amin(any_array).astype(np.float32)
        v_range = (vmax - vmin)
        any_array_nor = ((any_array - vmin) / v_range).astype(np.float16)
        self.sequence = 2 * (any_array_nor - 0.5)

        if log_dict != None:
            if type(log_dict) is dict:
                self.seq_log = log_dict
            else:
                raise ValueError, '`log_dict` should be a dictionary!'
        else:
            self.seq_log = {}
        self.clear()

    def set_stim(self, stim):
        """
        Calls the `generate_movie` method of the respective stim object and
        populates the attributes `self.sequence` and `self.seq_log`

        Parameters
        ----------
        stim : Stim object
            the type of stimulus to be presented in the experiment
        """
        if self.is_by_index:
            if stim.stim_name in ['KSstim', 'KSstimAllDir']:
                raise LookupError('Stimulus {} does not support indexed display.'.format(stim.name))

            self.sequence, self.seq_log = stim.generate_movie_by_index()
            self.clear()

        else:
            if stim.stim_name in ['LocallySparseNoise', 'StaticGratingCircle', 'NaturalScene']:
                raise LookupError('Stimulus {} does not support full sequence display. Please use '
                                  'indexed display instead (set self.is_by_index = True).')

            self.sequence, self.seq_log = stim.generate_movie()
            self.clear()

    def trigger_display(self):
        """
        Display stimulus, initialize and perform global experimental routines.

        Prepares all of the necessary parameters to display stimulus and store
        the data collected in the experiment. Interacts with PyschoPy to create
        and display each frame of the selected stimulus routine. Handles
        global calls to trigger and timing devices within the experimental setup.

        Examples
        --------
        >>> # Assume monitor, indicator, and stimulus objects are defined
        >>> import DisplaySequence
        >>> ds = DisplaySequence(log_dir=r'C:\)
        >>> ds.set_stim(uniform_contrast)
        >>> ds.trigger_display()

        """
        # --------------- early preparation for display--------------------
        # test monitor resolution
        try:
            resolution = self.seq_log['monitor']['resolution'][::-1]
        except KeyError:
            resolution = (800, 600)

        # test monitor refresh rate
        try:
            refresh_rate = self.seq_log['monitor']['refresh_rate']
        except KeyError:
            print("No monitor refresh rate information, assuming 60Hz.\n")
            refresh_rate = 60.

        # prepare display frames log
        if self.sequence is None:
            raise LookupError("Please set the sequence to be displayed by using self.set_stim().\n")
        if not self.seq_log:
            raise LookupError("Please set the sequence log dictionary to be displayed "
                              "by using self.set_stim().\n")

        # if display by index, check frame indices were not larger than the number of frames in
        # self.sequence
        if self.is_by_index:
            max_index = max(self.seq_log['stimulation']['index_to_display'])
            min_index = min(self.seq_log['stimulation']['index_to_display'])
            if max_index >= self.sequence.shape[0] or min_index < 0:
                raise ValueError('Max display index range: {} is out of self.sequence frame range: {}.'
                                 .format((min_index, max_index), (0, self.sequence.shape[0] - 1)))
            if 'frames_unique' not in self.seq_log['stimulation'].keys():
                raise LookupError('"frames_unique" is not found in self.seq_log["stimulation"]. This'
                                  'is required when display by index.')
        else:
            if 'frames' not in self.seq_log['stimulation'].keys():
                raise LookupError('"frames" is not found in self.seq_log["stimulation"]. This'
                                  'is required when display by full sequence.')

        # calculate expected display time
        if self.is_by_index:
            index_to_display = self.seq_log['stimulation']['index_to_display']
            display_time = (float(len(index_to_display)) *
                            self.display_iter / refresh_rate)
        else:
            display_time = (float(self.sequence.shape[0]) *
                            self.display_iter / refresh_rate)
        print('\nExpected display time: {} seconds.\n'.format(display_time))

        # generate file name
        self._get_file_name()
        print('File name: {}.\n'.format(self.file_name))

        # -----------------setup psychopy window and stimulus--------------
        # start psychopy window
        window = visual.Window(size=resolution,
                               monitor=self.psychopy_mon,
                               fullscr=True,
                               screen=self.display_screen,
                               color=self.initial_background_color)

        stim = visual.ImageStim(window, size=(2, 2), interpolate=self.is_interpolate)

        # initialize keep_display
        self.keep_display = True

        # handle display trigger
        if self.is_triggered:
            display_wait = self._wait_for_trigger(event=self.trigger_event)
            if not display_wait:
                window.close()
                self.clear()
                return None
            else:
                time.sleep(5.)  # wait remote object to start

        # actual display
        self._display(window=window, stim=stim)

        # analyze frames
        try:
            self.frame_duration, self.frame_stats = \
                analyze_frames(ts_start=self.frame_ts_start, ts_end=self.frame_ts_end,
                               refresh_rate=self.seq_log['monitor']['refresh_rate'])
        except KeyError:
            print("No monitor refresh rate information, assuming 60Hz.")
            self.frame_duration, self.frame_stats = \
                analyze_frames(ts_start=self.frame_ts_start, ts_end=self.frame_ts_end, refresh_rate=60.)

        self.save_log()

        # clear display data
        self.clear()

    def _wait_for_trigger(self, event):
        """
        time place holder for waiting for trigger

        Parameters
        ----------
        event : str from {'low_level','high_level','negative_edge','positive_edge'}
            an event triggered via a National Instuments experimental device.
        Returns
        -------
        Bool :
            returns `True` if trigger is detected and `False` if manual stop
            signal is detected.
        """

        # check NI signal
        trigger_task = iodaq.DigitalInput(self.trigger_NI_dev,
                                          self.trigger_NI_port,
                                          self.trigger_NI_line)
        trigger_task.StartTask()

        print("Waiting for trigger: {} on {}.".format(event, trigger_task.devstr))

        if event == 'low_level':
            last_TTL = trigger_task.read()
            while last_TTL != 0 and self.keep_display:
                last_TTL = trigger_task.read()[0]
                self._update_display_status()
            else:
                if self.keep_display:
                    trigger_task.StopTask()
                    print('Trigger detected. Start displaying...\n\n')
                    return True
                else:
                    trigger_task.StopTask()
                    print('Keyboard interrupting signal detected. Stopping the program.')
                    return False
        elif event == 'high_level':
            last_TTL = trigger_task.read()[0]
            while last_TTL != 1 and self.keep_display:
                last_TTL = trigger_task.read()[0]
                self._update_display_status()
            else:
                if self.keep_display:
                    trigger_task.StopTask()
                    print('Trigger detected. Start displaying...\n\n')
                    return True
                else:
                    trigger_task.StopTask()
                    print('Keyboard interrupting signal detected. Stopping the program.')
                    return False
        elif event == 'negative_edge':
            last_TTL = trigger_task.read()[0]
            while self.keep_display:
                current_TTL = trigger_task.read()[0]
                if (last_TTL == 1) and (current_TTL == 0):
                    break
                else:
                    last_TTL = int(current_TTL)
                    self._update_display_status()
            else:
                trigger_task.StopTask()
                print('Keyboard interrupting signal detected. Stopping the program.')
                return False
            trigger_task.StopTask()
            print('Trigger detected. Start displaying...\n\n')
            return True
        elif event == 'positive_edge':
            last_TTL = trigger_task.read()[0]
            while self.keep_display:
                current_TTL = trigger_task.read()[0]
                if (last_TTL == 0) and (current_TTL == 1):
                    break
                else:
                    last_TTL = int(current_TTL)
                    self._update_display_status()
            else:
                trigger_task.StopTask();
                print('Keyboard interrupting signal detected. Stopping the program.')
                return False
            trigger_task.StopTask()
            print('Trigger detected. Start displaying...\n\n')
            return True
        else:
            raise NameError, "`trigger` not in " \
                             "{'negative_edge','positive_edge', 'high_level','low_level'}!"

    def _get_file_name(self):
        """
        generate the file name of log file
        """

        try:
            self.file_name = datetime.datetime.now().strftime('%y%m%d%H%M%S') + \
                             '-' + self.seq_log['stimulation']['stim_name'] + \
                             '-M' + self.mouse_id + '-' + self.user_id + '-' + \
                             self.identifier
        except KeyError:
            self.file_name = datetime.datetime.now().strftime('%y%m%d%H%M%S') + \
                             '-' + 'customStim' + '-M' + self.mouse_id + '-' + \
                             self.user_id + '-' + self.identifier

        if self.is_triggered:
            self.file_name += '-Triggered'
        else:
            self.file_name += '-notTriggered'

    def _display(self, window, stim):
        """
        display stimulus
        """
        frame_ts_start = []
        frame_ts_end = []
        start_time = time.clock()

        if self.is_by_index:
            index_to_display = self.seq_log['stimulation']['index_to_display']
            iter_frame_num = len(index_to_display)
        else:
            iter_frame_num = self.sequence.shape[0]
            index_to_display = range(iter_frame_num)

        # print('frame per iter: {}'.format(iter_frame_num))

        if self.is_sync_pulse:
            syncPulseTask = iodaq.DigitalOutput(self.sync_pulse_NI_dev,
                                                self.sync_pulse_NI_port,
                                                self.sync_pulse_NI_line)
            syncPulseTask.StartTask()
            _ = syncPulseTask.write(np.array([0]).astype(np.uint8))

        i = 0
        self.displayed_frames = []

        while self.keep_display and i < (iter_frame_num * self.display_iter):

            frame_num = i % iter_frame_num
            frame_index = index_to_display[frame_num]

            # print('i:{}; index_display_ind:{}; frame_ind{}.'.format(i, frame_num, frame_index))

            if self.color_weights == (1., 1., 1.):
                stim.setImage(self.sequence[frame_index][::-1])
            else:
                curr_frame = self.sequence[frame_index]
                curr_frame = ((curr_frame + 1.) * 255 / 2.)
                curr_frame_r = PIL.Image.fromarray((curr_frame * self.color_weights[0]).astype(np.uint8))
                curr_frame_g = PIL.Image.fromarray((curr_frame * self.color_weights[1]).astype(np.uint8))
                curr_frame_b = PIL.Image.fromarray((curr_frame * self.color_weights[2]).astype(np.uint8))
                curr_frame = PIL.Image.merge('RGB', (curr_frame_r, curr_frame_g, curr_frame_b))
                # plt.imshow(curr_frame)
                # plt.show()
                stim.setImage(curr_frame)

            stim.draw()

            # set sync pulse start signal
            if self.is_sync_pulse:
                _ = syncPulseTask.write(np.array([1]).astype(np.uint8))

            # save frame start timestamp
            frame_ts_start.append(time.clock() - start_time)

            # show visual stim
            window.flip()

            # save displayed frame information
            if self.is_by_index:
                self.displayed_frames.append(self.seq_log['stimulation']['frames_unique'][frame_index])
            else:
                self.displayed_frames.append(self.seq_log['stimulation']['frames'][frame_index])

            # save frame end timestamp
            frame_ts_end.append(time.clock() - start_time)

            # set sync pulse end signal
            if self.is_sync_pulse:
                _ = syncPulseTask.write(np.array([0]).astype(np.uint8))

            self._update_display_status()
            i += 1

        stop_time = time.clock()
        window.close()

        if self.is_sync_pulse:
            syncPulseTask.StopTask()

        self.frame_ts_start = np.array(frame_ts_start)
        self.frame_ts_end = np.array(frame_ts_end)
        self.display_length = stop_time - start_time

        if self.keep_display == True:
            print('\nDisplay successfully completed.')

    def flag_to_close(self):
        self.keep_display = False

    def _update_display_status(self):

        if self.keep_display is None:
            raise LookupError, 'self.keep_display should start as True'

        # check keyboard input 'q' or 'escape'
        keyList = event.getKeys(['q', 'escape'])
        if len(keyList) > 0:
            self.keep_display = False
            print("Keyboard interrupting signal detected. Stop displaying. \n")

    def set_display_iteration(self, display_iter):

        if display_iter % 1 == 0:
            self.display_iter = display_iter
        else:
            raise ArithmeticError, "`display_iter` should be a whole number."
        self.clear()

    def save_log(self):

        if self.display_length is None:
            self.clear()
            raise LookupError("Please display sequence first!")

        if self.file_name is None:
            self._get_file_name()

        if self.keep_display == True:
            self.file_name += '-complete'
        elif self.keep_display == False:
            self.file_name += '-incomplete'

        # set up log object
        directory = os.path.join(self.log_dir, 'visual_display_log')
        if not (os.path.isdir(directory)):
            os.makedirs(directory)

        logFile = dict(self.seq_log)
        displayLog = dict(self.__dict__)
        displayLog.pop('seq_log')
        displayLog.pop('sequence')
        logFile.update({'presentation': displayLog})

        file_name = self.file_name + ".pkl"

        # generate full log dictionary
        path = os.path.join(directory, file_name)
        ft.saveFile(path, logFile)
        if self.is_save_sequence:
            tf.imsave(os.path.join(directory, self.file_name + '.tif'),
                      self.sequence.astype(np.float32))
        print(".pkl file generated successfully.")

        backupFileFolder = self._get_backup_folder()
        if backupFileFolder is not None:
            if not (os.path.isdir(backupFileFolder)):
                os.makedirs(backupFileFolder)
            backupFilePath = os.path.join(backupFileFolder, file_name)
            ft.saveFile(backupFilePath, logFile)

            if self.is_save_sequence:
                tf.imsave(os.path.join(backupFileFolder, self.file_name + '.tif'),
                          self.sequence.astype(np.float32))
            print(".pkl backup file generate successfully")
        else:
            print("Did not find backup path, no backup has been saved.")

    def _get_backup_folder(self):
        if self.backupdir is not None:
            backup_folder = os.path.join(self.backupdir, 'visual_display_log')
            if not os.path.isdir(backup_folder):
                os.makedirs(backup_folder)
            return backup_folder
        else:
            return None

    def clear(self):
        """ clear display information. """
        self.display_length = None
        self.time_stamp = None
        self.frame_duration = None
        self.displayed_frames = None
        self.frame_stats = None
        self.file_name = None
        self.keep_display = None


if __name__ == "__main__":
    pass



__author__ = 'junz'

import matplotlib.pyplot as plt
import numpy as np
from scipy import interpolate
import scipy.ndimage as ni
import skimage.morphology as sm
import FileTools as ft
import PlottingTools as pt

try:
    import cv2
except ImportError as e:
    print e


def resample(t1, y1, interval, kind='linear', isPlot=False):
    """
    :param t1: time stamps of original data
    :param y1: value of original data (relative to t1)
    :param interval: the intervals of resampled time stamps, second
    :param kind: interpolation type, same as 'scipy.interpolate.interp1d'
    :return: t2, y2
    """

    f = interpolate.interp1d(t1, y1, kind=kind)

    t2 = np.arange(t1[0], t1[-1], interval)

    y2 = f(t2)

    if isPlot:
        plt.figure()
        plt.plot(t1, y1)
        plt.plot(t2, y2)
        plt.legend('original', 'resampled')

    return t2, y2


def resample2(t1, y1, t2, kind='linear', isPlot=False, bounds_error=False):
    """
    :param t1: time stamps of original data
    :param y1: value of original data (relative to t1)
    :param t2: time stamps for resample
    :param kind: interpolation type, same as 'scipy.interpolate.interp1d'
    :return: y2
    """

    f = interpolate.interp1d(t1, y1, kind=kind, bounds_error=bounds_error)

    y2 = f(t2)

    if isPlot:
        plt.figure()
        plt.plot(t1, y1)
        plt.plot(t2, y2)
        plt.legend('original', 'resampled')

    return y2


def array_nor(A):
    """
    normalize a np.array to the scale [0, 1]
    """

    B = A.astype(np.float)
    return (B - np.amin(B)) / (np.amax(B) - np.amin(B))


def array_nor_median(A):
    """
    normalize array by minus median, data type will be switch to np.float
    """
    B = A.astype(np.float)
    return B - np.median(B.flatten())


def array_nor_mean(A):
    """
    normalize array by minus mean, data type will be switch to np.float
    """
    B = A.astype(np.float)
    return B - np.mean(B.flatten())


def array_nor_mean_std(A):
    """
    normalize array by minus mean and then devided by standard deviation,
    data type will be switched to np.float
    """
    A = A.astype(np.float)
    B = (A - np.mean(A.flatten())) / np.std(A.flatten())
    return B


def zscore(A):
    """
    return Z score of an array.
    """

    if np.isnan(A).any():
        return (A - np.nanmean(A.flatten())) / np.nanstd(A.flatten())
    else:
        A = A.astype(np.float); return (A - np.mean(A.flatten())) / np.std(A.flatten())


def distance(p0, p1):
    """
    calculate distance between two points, can be multi-dimensinal

    p0 and p1 should be a 1d array, with each element for each dimension
    """

    if not isinstance(p0, np.ndarray): p0 = np.array(p0)
    if not isinstance(p1, np.ndarray): p1 = np.array(p1)
    return np.sqrt(np.sum(np.square(p0 - p1).flatten()))


def array_diff(a0, a1):
    """
    calculate the sum of pixel-wise difference between two arrays
    """
    if not isinstance(a0, np.ndarray): a0 = np.array(a0)
    if not isinstance(a1, np.ndarray): a1 = np.array(a1)
    return np.mean(np.abs(a0 - a1).flatten())


def binarize(array, threshold):
    """
    binarize array to 0s and 1s, by cutting at threshold
    """

    newArray = np.array(array)

    newArray[array >= threshold] = 1.

    newArray[array < threshold] = 0.

    newArray = newArray.astype(array.dtype)

    return newArray


def center_image(img,  # original image, 2d ndarray
                 centerPixel,  # the coordinates of center pixel in original image, [col, row]
                 newSize=512,  # the size of output image
                 borderValue=0
                 ):
    """
    center a certain image in a new canvas

    the pixel defined by 'centerPixel' in the original image will be at the center of the output image
    the size of output image is defined by 'newSize'

    empty pixels will be filled with zeros
    """

    x = newSize / 2 - centerPixel[1]
    y = newSize / 2 - centerPixel[0]

    M = np.float32([[1, 0, x], [0, 1, y]])

    newImg = cv2.warpAffine(img, M, (newSize, newSize), borderValue=borderValue)

    return newImg


def resize_image(img, outputShape, fillValue=0.):
    """
    resize every frame of a 3-d matrix to defined output shape
    if the original image is too big it will be truncated
    if the original image is too small, value defined as fillValue will filled in. default: 0
    """

    width = outputShape[1]
    height = outputShape[0]

    if width < 1:
        raise ValueError, 'width should be bigger than 0!!'

    if height < 1:
        raise ValueError, 'height should be bigger than 0!!'

    if len(img.shape) != 2 and len(img.shape) != 3:
        raise ValueError, 'input image should be a 2-d or 3-d array!!'

    if len(img.shape) == 2:  # 2-d image
        startWidth = img.shape[-1]
        startHeight = img.shape[-2]
        newImg = np.array(img)
        if startWidth > width:
            newImg = newImg[:, 0:width]
        elif startWidth < width:
            attachRight = np.zeros((startHeight, width - startWidth))
            attachRight[:] = fillValue
            attachRight.astype(img.dtype)
            newImg = np.hstack((newImg, attachRight))

        if startHeight > height:
            newImg = newImg[0:height, :]
        elif startHeight < height:
            attachBottom = np.zeros((height - startHeight, width))
            attachBottom[:] = fillValue
            attachBottom.astype(img.dtype)
            newImg = np.vstack((newImg, attachBottom))

    if len(img.shape) == 3:  # 3-d matrix
        startDepth = img.shape[0]
        startWidth = img.shape[-1]
        startHeight = img.shape[-2]
        newImg = np.array(img)
        if startWidth > width:
            newImg = newImg[:, :, 0:width]
        elif startWidth < width:
            attachRight = np.zeros((startDepth, startHeight, width - startWidth))
            attachRight[:] = fillValue
            attachRight.astype(img.dtype)
            newImg = np.concatenate((img, attachRight), axis=2)

        if startHeight > height:
            newImg = newImg[:, 0:height, :]
        elif startHeight < height:
            attachBottom = np.zeros((startDepth, height - startHeight, width))
            attachBottom[:] = fillValue
            attachBottom.astype(img.dtype)
            newImg = np.concatenate((newImg, attachBottom), axis=1)

    return newImg


def expand_image_cv2(img):
    if len(img.shape) != 2:
        raise ValueError, 'Input image should be 2d!'

    dtype = img.dtype
    img = img.astype(np.float32)
    rows, cols = img.shape
    diagonal = int(np.sqrt(rows ** 2 + cols ** 2))
    M = np.float32([[1, 0, (diagonal - cols) / 2], [0, 1, (diagonal - rows) / 2]])
    newImg = cv2.warpAffine(img, M, (diagonal, diagonal))
    return newImg.astype(dtype)


def expand_image(img):
    if len(img.shape) == 2:
        rows, cols = img.shape
        diagonal = int(np.sqrt(rows ** 2 + cols ** 2))
        top = np.zeros(((diagonal - rows) / 2, cols), dtype=img.dtype)
        down = np.zeros((diagonal - img.shape[0] - top.shape[0], cols), dtype=img.dtype)
        tall = np.vstack((top, img, down))
        left = np.zeros((tall.shape[0], (diagonal - cols) / 2), dtype=img.dtype)
        right = np.zeros((tall.shape[0], diagonal - img.shape[1] - left.shape[1]), dtype=img.dtype)
        newImg = np.hstack((left, tall, right))
        return newImg
    elif len(img.shape) == 3:
        frames, rows, cols = img.shape
        diagonal = int(np.sqrt(rows ** 2 + cols ** 2))
        top = np.zeros((frames, (diagonal - rows) / 2, cols), dtype=img.dtype)
        down = np.zeros((frames, diagonal - img.shape[1] - top.shape[1], cols), dtype=img.dtype)
        tall = np.concatenate((top, img, down), axis=1)
        left = np.zeros((frames, tall.shape[1], (diagonal - cols) / 2), dtype=img.dtype)
        right = np.zeros((frames, tall.shape[1], diagonal - img.shape[2] - left.shape[2]), dtype=img.dtype)
        newImg = np.concatenate((left, tall, right), axis=2)
        return newImg
    else:
        raise ValueError, 'Input image should be 2d or 3d!'


def zoom_image(img, zoom, interpolation='cubic'):  # 'cubic','linear','area','nearest','lanczos4'
    """
    zoom a 2d image. if zoom is a single value, it will apply to both axes, if zoom has two values it will be applied to
    height and width respectively
    zoom[0]: height
    zoom[1]: width
    """
    if len(img.shape) != 2:
        raise ValueError, 'Input image should be 2d!'

    try:
        zoomH = float(zoom[0]); zoomW = float(zoom[1])
    except TypeError:
        zoomH = float(zoom); zoomW = float(zoom)

    if interpolation == 'cubic': interpo = cv2.INTER_CUBIC
    if interpolation == 'linear': interpo = cv2.INTER_LINEAR
    if interpolation == 'area': interpo = cv2.INTER_AREA
    if interpolation == 'nearest': interpo = cv2.INTER_NEAREST
    if interpolation == 'lanczos4': interpo = cv2.INTER_LANCZOS4

    newImg = cv2.resize(img.astype(np.float), dsize=(int(img.shape[1] * zoomW), int(img.shape[0] * zoomH)),
                        interpolation=interpo)
    return newImg


def moveImage(img, Xoffset, Yoffset, width, height, borderValue=0.0):
    """
    move image defined by Xoffset and Yoffset

    new canvas size is defined by width and height

    empty pixels will be filled with zeros
    """
    if len(img.shape) != 2:
        raise ValueError, 'Input image should be 2d!'

    M = np.float32([[1, 0, Xoffset], [0, 1, Yoffset]])

    newImg = cv2.warpAffine(img, M, (width, height), borderValue=borderValue)

    return newImg


def rotate_image(img, angle, borderValue=0.0):
    """
    rotate an image conterclock wise by an angle defined by 'angle' in degree

    pixels go out side will be tropped
    pixels with no value will be filled as zeros
    """

    if len(img.shape) != 2:
        raise ValueError, 'Input image should be 2d!'

    rows, cols = img.shape

    M = cv2.getRotationMatrix2D((cols / 2, rows / 2), angle, 1)
    # M = cv2.getRotationMatrix2D((0,0),angle,1)
    newImg = cv2.warpAffine(img, M, (cols, rows), borderValue=borderValue)

    return newImg


def rigid_transform(img, zoom=None, rotation=None, offset=None, outputShape=None, mode='constant', cval=0.0):
    """
    rigid transformation of a 2d-image or 3d-matrix by using scipy
    :param img: input image/matrix
    :param zoom:
    :param rotation: in degree, counterclock wise
    :param offset: tuple (xoffset, yoffset) pixel value of starting point of output image
    :param outputShape: the shape of output image, (height, width)
    :return: new image or matrix after transformation
    """

    if len(img.shape) != 2 and len(img.shape) != 3:
        raise LookupError, 'Input image is not a 2d or 3d array!'

    newImg = img.astype(np.float32)

    if zoom:
        if len(img.shape) == 2:
            newZoom = (zoom, zoom)
        elif len(img.shape) == 3:
            newZoom = (1, zoom, zoom)
        newImg = ni.zoom(newImg, zoom=newZoom, mode=mode, cval=cval)

    if rotation:
        newImg = expand_image(newImg)
        if len(img.shape) == 2:
            newImg = ni.rotate(newImg, angle=rotation, reshape=False, mode=mode, cval=cval)
        elif len(img.shape) == 3:
            newImg = ni.rotate(newImg, angle=rotation, axes=(1, 2), reshape=False, mode=mode, cval=cval)

    if offset:
        if len(img.shape) == 2:
            newImg = ni.shift(newImg, (offset[1], offset[0]), mode=mode, cval=cval)
        if len(img.shape) == 3:
            newImg = ni.shift(newImg, (0, offset[1], offset[0]), mode=mode, cval=cval)

    if outputShape:
        newImg = resize_image(newImg, outputShape)

    return newImg.astype(img.dtype)


def rigid_transform_cv2_2d(img, zoom=None, rotation=None, offset=None, outputShape=None):
    """
    rigid transformation of a 2d-image by using opencv
    :param img: input image/matrix
    :param zoom:
    :param rotation: in degree, counterclock wise
    :param offset: tuple (xoffset, yoffset) pixel value of starting point of output image
    :param outputShape: the shape of output image, (height, width)
    :return: new image or matrix after transformation
    """

    if len(img.shape) != 2:
        raise LookupError, 'Input image is not a 2d or 3d array!'

    newImg = np.array(img).astype(np.float)
    minValue = np.amin(newImg)

    if zoom:
        newImg = zoom_image(img, zoom=zoom)

    if rotation:
        newImg = expand_image_cv2(newImg)
        newImg = rotate_image(newImg, rotation, borderValue=minValue)

    if (outputShape is None) and (offset is None):
        return newImg
    else:
        if outputShape is None:
            outputShape = newImg.shape
        if offset is None:
            offset = (0, 0)
        newImg = moveImage(newImg, offset[0], offset[1], outputShape[1], outputShape[0], borderValue=minValue)

        return newImg.astype(img.dtype)


def rigid_transform_cv2_3d(img, zoom=None, rotation=None, offset=None, outputShape=None):
    if len(img.shape) != 3:
        raise LookupError, 'Input image is not a 3d array!'

    if not outputShape:
        if zoom:
            newHeight = int(img.shape[1] * zoom)
            newWidth = int(img.shape[2] * zoom)
        else:
            newHeight = img.shape[1]
            newWidth = img.shape[2]
    else:
        newHeight = outputShape[0]
        newWidth = outputShape[1]
    newImg = np.empty((img.shape[0], newHeight, newWidth), dtype=img.dtype)

    for i in range(img.shape[0]):
        newImg[i, :, :] = rigid_transform_cv2_2d(img[i, :, :], zoom=zoom, rotation=rotation, offset=offset,
                                                 outputShape=outputShape)

    return newImg


def rigid_transform_cv2(img, zoom=None, rotation=None, offset=None, outputShape=None):
    """
    rigid transformation of a 2d-image or 3d-matrix by using opencv
    :param img: input image/matrix
    :param zoom:
    :param rotation: in degree, counterclock wise
    :param offset: tuple (xoffset, yoffset) pixel value of starting point of output image
    :param outputShape: the shape of output image, (height, width)
    :return: new image or matrix after transformation
    """

    if len(img.shape) == 2:
        return rigid_transform_cv2_2d(img, zoom=zoom, rotation=rotation, offset=offset, outputShape=outputShape)
    elif len(img.shape) == 3:
        return rigid_transform_cv2_3d(img, zoom=zoom, rotation=rotation, offset=offset, outputShape=outputShape)
    else:
        raise ValueError, 'Input image is not a 2d or 3d array!'


def boxcartime_dff(data,
                   window,  # boxcar size in seconds
                   fs  # sample rate in ms
                   ):
    """
    Created on Mon Nov 24 14:37:02 2014

    [dff] = boxcartime_dff(data[t,y,x], rollingwindow[in s], samplerate[in ms])
    boxcar average uses scipy.signal package, imported locally

    @author: mattv
    """
    import scipy.signal as sig

    if data.ndim != 3:
        raise LookupError, 'input images must be a 3-dim array format [t,y,x]'

    exposure = np.float(fs / 1000)  # convert exposure from ms to s
    win = np.float(window)
    win = win / exposure
    win = np.ceil(win)

    # rolling average
    kernal = np.ones(win, dtype=('float'))
    padsize = data.shape[0] + win * 2
    mov_ave = np.zeros([padsize + win - 1], dtype=('float'))
    mov_pad = np.zeros([padsize], dtype=('float'))
    mov_dff = np.zeros([data.shape[0] - win, data.shape[1], data.shape[2]])
    for y in range(data.shape[1]):
        for x in range(data.shape[2]):
            # put data within padded array
            mov_pad[win:(padsize - win)] = data[:, y, x]
            # moving average by convolution
            mov_ave = sig.fftconvolve(mov_pad, kernal) / win
            # cut off pad
            mov_ave = mov_ave[win * 2:1 + mov_ave.shape[0] - win * 2]
            # use moving average as f0 for df/f
            mov_dff[:, y, x] = (data[(win / 2):data.shape[0] - (win / 2), y, x] - mov_ave) / mov_ave

    return mov_dff


def normalize_movie(movie,
                    baselinePic=None,  # picture for baseline
                    baselineType='mean'  # 'mean' or 'median'
                    ):
    """
    return average image, movie minus avearage, and dF over F for each pixel
    """
    movie = np.array(movie, dtype=np.float32)

    if baselinePic is not None:

        if movie.shape[1:] != baselinePic.shape:
            raise LookupError, 'The shape of "baselinePic" should match the shape of the frame shape of "movie"!'

        averageImage = baselinePic

    elif baselineType == 'mean':
        averageImage = np.mean(movie, axis=0)

    elif baselineType == 'median':
        averageImage = np.median(movie, axis=0)

    else:
        raise LookupError, 'The "baselineType" should be "mean" or "median"!!'

    normalizedMovie = np.subtract(movie, averageImage)
    dFoverFMovie = np.divide(normalizedMovie, averageImage)

    return averageImage, normalizedMovie, dFoverFMovie


def temporal_filter_movie(mov,  # array of movie
                          Fs,  # sampling rate
                          Flow,  # low cutoff frequency
                          Fhigh,  # high cutoff frequency
                          mode='box'):  # filter mode, '1/f' or 'box'):

    if len(mov.shape) != 3:
        raise LookupError, 'The "mov" array should have 3 dimensions!'

    frameNum = mov.shape[0]
    freqs = np.fft.fftfreq(frameNum, d=(1. / float(Fs)))

    filterArray = np.ones(frameNum)

    for i in xrange(frameNum):
        if ((freqs[i] > 0) and (freqs[i] < Flow) or (freqs[i] > Fhigh)) or \
                ((freqs[i] < 0) and (freqs[i] > -Flow) or (freqs[i] < -Fhigh)):
            filterArray[i] = 0

    if mode == '1/f':
        filterArray[1:] = filterArray[1:] / abs(freqs[1:])
        filterArray[0] = 0
        filterArray = (filterArray - np.amin(filterArray)) / (np.amax(filterArray) - np.amin(filterArray))
    elif mode == 'box':
        filterArray[0] = 0
    else:
        raise NameError, 'Variable "mode" should be either "1/f" or "box"!'

    if Flow == 0:
        filterArray[0] = 1

    movFFT = np.fft.fft(mov, axis=0)

    for i in xrange(mov.shape[1]):
        for j in xrange(mov.shape[2]):
            movFFT[:, i, j] = movFFT[:, i, j] * filterArray

    movF = np.real(np.fft.ifft(movFFT, axis=0))

    return movF


def generate_rectangle_mask(shape, center, width, height, isplot=False):
    if len(shape) != 2: raise LookupError, 'Shape should be two dimensional.'

    mask = np.zeros(shape);
    mask[:] = np.nan
    mask[int(round(center[0] - height / 2)):int(round(center[0] + height / 2)),
    int(round(center[1] - width / 2)):int(round(center[1] + width / 2))] = 1

    if np.isnan(np.nansum(mask[:])):
        raise ArithmeticError, 'No element in mask!'

    if isplot == True:
        f = plt.figure();
        ax = f.add_subplot(111)
        pt.plot_mask_borders(mask, plotAxis=ax)

    return mask


def generate_oval_mask(shape, center, width, height, isplot=False):
    if len(shape) != 2: raise LookupError, 'Shape should be two dimensional.'

    mask = np.zeros(shape);
    mask[:] = np.nan

    width = float(width);
    height = float(height)

    for i in range(shape[0]):
        for j in range(shape[1]):
            if ((i - center[0]) / (height / 2)) ** 2 + ((j - center[1]) / (width / 2)) ** 2 <= 1:
                mask[i, j] = 1

    if np.isnan(np.nansum(mask[:])):
        raise ArithmeticError, 'No element in mask!'

    if isplot == True:
        f = plt.figure();
        ax = f.add_subplot(111)
        pt.plot_mask_borders(mask, plotAxis=ax)

    return mask


def get_trace(movie, mask, maskMode='binary'):
    """
    get a trace across a movie with averaged value in a mask

    maskMode: 'binary': ones in roi, zeros outside
              'binaryNan': ones in roi, nans outside
              'weighted': weighted values in roi, zeros outside (note: all pixels equal to zero will be considered outside roi
              'weightedNan': weighted values in roi, nans outside
    """

    if maskMode == 'binary':
        if np.where(mask == 0)[0].size + np.where(mask == 1)[0].size < mask.size:
            raise ValueError, 'Binary mask should only contain zeros and ones!!'
        else:
            finalMask = np.array(mask.astype(np.float))
            pixelNum = np.sum(finalMask.flatten())
    elif maskMode == 'binaryNan':
        if np.sum(np.isnan(mask).flatten()) + np.where(mask == 1)[0].size < mask.size:
            raise ValueError, 'BinaryNan mask should only contain nans and ones!!'
        else:
            finalMask = np.ones(mask.shape, dtype=np.float)
            finalMask[np.isnan(mask)] = 0
            pixelNum = mask.size - np.sum(np.isnan(mask).flatten())
    elif maskMode == 'weighted':
        if np.isnan(mask).any():
            raise ValueError, 'Weighted mask should not contain nan(s)!!'
        else:
            finalMask = np.array(mask.astype(np.float))
            pixelNum = mask.size - np.where(mask == 0)[0].size
    elif maskMode == 'weightedNan':
        finalMask = np.array(mask.astype(np.float))
        finalMask[np.isnan(mask)] = 0
        pixelNum = mask.size - np.where(finalMask == 0)[0].size
    else:
        raise LookupError, 'maskMode not understood. Should be one of "binary", "binaryNan", "weighted", "weightedNan".'

    trace = np.sum(np.multiply(movie, finalMask), (1, 2)) / pixelNum

    return trace


def get_trace_binaryslicer(bl_obj, mask, mask_mode='binary'):
    """

    :param bl_obj: the binary slicer object of a large matrix
    :param mask: the mask
    :param mask_mode: same as 'mask_mode' in function get_trace

    maskMode: 'binary': ones in roi, zeros outside
              'binaryNan': ones in roi, nans outside
              'weighted': weighted values in roi, zeros outside (note: all pixels equal to zero will be considered outside roi
              'weightedNan': weighted values in roi, nans outside

    :return: extracted trace
    """

    if len(bl_obj.shape) != 3: raise ValueError, 'BinarySlicer object should be 3d!'
    if len(mask.shape) != 2: raise ValueError, 'Mask should be 2d!'
    if bl_obj.shape[1] != mask.shape[0] or bl_obj.shape[2] != mask.shape[1]:
        raise ValueError, 'the size of each frame of the BinarySlicer object should be the same as the size of mask'

    if mask_mode == 'binary':
        if np.where(mask == 0)[0].size + np.where(mask == 1)[0].size < mask.size:
            raise ValueError, 'Binary mask should only contain zeros and ones!!'
        else:
            mask_ind = np.where(mask != 0)
            # print mask_ind
            min_row = min(mask_ind[0]);
            max_row = max(mask_ind[0]) + 1
            min_col = min(mask_ind[1]);
            max_col = max(mask_ind[1]) + 1
            finalMask = np.array(mask.astype(np.float))[min_row:max_row, min_col:max_col]
    elif mask_mode == 'binaryNan':
        if np.sum(np.isnan(mask).flatten()) + np.where(mask == 1)[0].size < mask.size:
            raise ValueError, 'BinaryNan mask should only contain nans and ones!!'
        else:
            mask_ind = np.where(mask != np.nan)
            min_row = min(mask_ind[0]);
            max_row = max(mask_ind[0]) + 1
            min_col = min(mask_ind[1]);
            max_col = max(mask_ind[1]) + 1
            finalMask = np.ones(mask.shape, dtype=np.float)
            finalMask[np.isnan(mask)] = 0
            finalMask = finalMask[min_row:max_row, min_col:max_col]
    elif mask_mode == 'weighted':
        if np.isnan(mask).any():
            raise ValueError, 'Weighted mask should not contain nan(s)!!'
        else:
            mask_ind = np.where(mask != 0)
            min_row = min(mask_ind[0]);
            max_row = max(mask_ind[0]) + 1
            min_col = min(mask_ind[1]);
            max_col = max(mask_ind[1]) + 1
            finalMask = np.array(mask.astype(np.float))[min_row:max_row, min_col:max_col]
    elif mask_mode == 'weightedNan':
        finalMask = np.array(mask.astype(np.float))
        finalMask[np.isnan(mask)] = 0
        mask_ind = np.where(finalMask != 0)
        min_row = min(mask_ind[0]);
        max_row = max(mask_ind[0]) + 1
        min_col = min(mask_ind[1]);
        max_col = max(mask_ind[1]) + 1
        finalMask = finalMask[min_row:max_row, min_col:max_col]

    mov = bl_obj[:, min_row:max_row, min_col:max_col]
    # print mov
    return get_trace(mov, finalMask, maskMode='weighted')


def get_trace_binaryslicer2(bl_obj, mask, mask_mode='binary', loading_frame_num=1000):
    """

    get trace for a given mask from a BinarySlicer object, by loading chunk each time

    :param bl_obj: the binary slicer object of a large matrix
    :param mask: the mask
    :param mask_mode: same as 'mask_mode' in function get_trace
    :param loading_frame_num: frame number of each chunk

    maskMode: 'binary': ones in roi, zeros outside
              'binaryNan': ones in roi, nans outside
              'weighted': weighted values in roi, zeros outside (note: all pixels equal to zero will be considered outside roi
              'weightedNan': weighted values in roi, nans outside

    :return: extracted trace
    """

    if loading_frame_num <= 1: raise ValueError, 'loading_frame_num should be a integer larger than 1!'
    if len(bl_obj.shape) != 3: raise ValueError, 'BinarySlicer object should be 3d!'
    if len(mask.shape) != 2: raise ValueError, 'Mask should be 2d!'
    if bl_obj.shape[1] != mask.shape[0] or bl_obj.shape[2] != mask.shape[1]:
        raise ValueError, 'the size of each frame of the BinarySlicer object should be the same as the size of mask'

    frameNum = bl_obj.shape[0]

    print '\nInput movie shape:', bl_obj.shape

    chunkNum = frameNum // loading_frame_num
    if frameNum % loading_frame_num == 0:
        print 'Translating in chunks: ' + str(chunkNum) + ' x ' + str(loading_frame_num) + ' frame(s)'
    else:
        chunkNum += 1
        print 'Translating in chunks: ' + str(chunkNum - 1) + ' x ' + str(
            loading_frame_num) + ' frame(s)' + ' + ' + str(frameNum % loading_frame_num) + ' frame(s)'

    traces = []
    for i in range(chunkNum):
        indStart = i * loading_frame_num
        indEnd = (i + 1) * loading_frame_num
        if indEnd > frameNum: indEnd = frameNum
        print 'Extracting signal from frame ' + str(indStart) + ' to frame ' + str(indEnd) + '.\t' + str(
            i * 100. / chunkNum) + '%'
        currMov = bl_obj[indStart:indEnd, :, :]
        traces.append(get_trace(currMov, mask, maskMode=mask_mode))

    return np.concatenate(traces)


def get_trace_binaryslicer3(bl_obj, masks, mask_mode='binary', loading_frame_num=1000):
    """

    get trace for a given mask from a BinarySlicer object, by loading chunk each time

    :param bl_obj: the binary slicer object of a large matrix
    :param masks: a dictionary of masks
    :param mask_mode: same as 'mask_mode' in function get_trace
    :param loading_frame_num: frame number of each chunk

    maskMode: 'binary': ones in roi, zeros outside
              'binaryNan': ones in roi, nans outside
              'weighted': weighted values in roi, zeros outside (note: all pixels equal to zero will be considered outside roi
              'weightedNan': weighted values in roi, nans outside

    :return: extracted trace
    """

    if loading_frame_num <= 1: raise ValueError, 'loading_frame_num should be a integer larger than 1!'
    if len(bl_obj.shape) != 3: raise ValueError, 'BinarySlicer object should be 3d!'

    frameNum = bl_obj.shape[0]

    print '\nInput movie shape:', bl_obj.shape

    chunkNum = frameNum // loading_frame_num
    if frameNum % loading_frame_num == 0:
        print 'Translating in chunks: ' + str(chunkNum) + ' x ' + str(loading_frame_num) + ' frame(s)'
    else:
        chunkNum += 1
        print 'Translating in chunks: ' + str(chunkNum - 1) + ' x ' + str(
            loading_frame_num) + ' frame(s)' + ' + ' + str(frameNum % loading_frame_num) + ' frame(s)'

    traces = {}
    for key in masks.iterkeys(): traces.update({'trace_' + key: []})

    for i in range(chunkNum):
        indStart = i * loading_frame_num
        indEnd = (i + 1) * loading_frame_num
        if indEnd > frameNum: indEnd = frameNum
        print 'Extracting signal from frame ' + str(indStart) + ' to frame ' + str(indEnd) + '.\t' + str(
            i * 100. / chunkNum) + '%'
        currMov = bl_obj[indStart:indEnd, :, :]
        for key, mask in masks.iteritems():
            if len(mask.shape) != 2: raise ValueError, 'Mask "' + key + '" should be 2d!'
            if bl_obj.shape[1] != mask.shape[0] or bl_obj.shape[2] != mask.shape[1]:
                raise ValueError, 'the size of each frame of the BinarySlicer object should be the same as the size of mask "' + key + '"!'
            traces['trace_' + key].append(get_trace(currMov, mask, maskMode=mask_mode))

    for key in traces.iterkeys():
        traces[key] = np.concatenate(traces[key])

    return traces


def hit_or_miss(coor, mask):
    """
    check if a cooridnate (coor) is in a mask, input mask can be int or float, nan and zero will considered as outside, any
    non-nan, non-zero pixel will be considered as inside. Mask does not need to be continuous.
    """
    mask[np.isnan(mask)] = 0;
    mask[mask > 0] = 1
    mask = mask.astype(np.int8)
    corMask = np.zeros(mask.shape, dtype=np.int8)
    corMask[np.round(coor[0]), np.round(coor[1])] = 1
    if np.sum(np.multiply(corMask, mask)) > 0:
        return True
    if np.sum(np.multiply(corMask, mask)) == 0:
        return False


def harmonic_amplitude(f,  # function value
                       period,  # how many fundamental harmonic periods inside the function
                       n):  # return the n-th harmonic
    """
    calculate the amplitude and phase of the n-th harmonic components of a
    function. the input function should have whole number times of period of
    the fundamental harmonic.
    """

    if (type(period) != int) | (period <= 0):
        raise ArithmeticError, '"period" should be a positive integer!'

    if (type(n) != int) | (n < 0):
        raise ArithmeticError, '"n" should be a non-negative positive integer!'

    L = len(f)
    x = np.arange(L)

    if n == 0:
        har = np.sum(np.multiply(f, np.exp(-1j * 2 * np.pi * period * n * x / L))) / L
        har_meg = np.abs(har)
        har_phase = np.nan

    else:
        har = 2 * np.sum(np.multiply(f, np.exp(-1j * 2 * np.pi * period * n * x / L))) / L
        har_meg = np.abs(har)
        har_phase = np.angle(har)

    return har_meg, har_phase


def discretize(array, binSize):
    """
    discretize the array by binSize
    """

    bins = np.arange(np.floor(np.nanmin(array)) - (0.5 * binSize),
                     np.ceil(np.nanmax(array)) + (1.5 * binSize),
                     binSize)

    flatArray = np.ndarray.flatten(array)

    indArray = np.digitize(flatArray, bins)

    newArray = np.zeros(flatArray.shape)
    newArray[:] = np.nan

    for i in xrange(len(indArray)):
        if np.isnan(flatArray[i]) == False:
            newArray[i] = bins[indArray[i]]

    newArray = np.array(newArray).reshape(array.shape)

    return newArray


def seed_pixel(markers):
    """
    marker centroid of every marked local minimum
    """

    newMarkers = np.zeros(markers.shape).astype(np.int32)
    intMarkers = markers.astype(np.int32)

    for i in range(1, np.amax(intMarkers) + 1):
        aa = np.zeros(markers.shape).astype(np.int32)
        aa[intMarkers == i] = 1
        aapixels = np.argwhere(aa)
        center = np.mean(aapixels.astype(np.float32), axis=0)

        newMarkers[int(np.round(center[0])), int(np.round(center[1]))] = i

    return newMarkers


def is_adjacent(array1, array2, borderWidth=2):
    """
    decide if two patches are adjacent within border width
    """

    p1d = ni.binary_dilation(array1, iterations=borderWidth - 1).astype(np.int8)
    p2d = ni.binary_dilation(array2, iterations=borderWidth - 1).astype(np.int8)

    if np.amax(p1d + p2d) > 1:
        return True
    else:
        return False


def remove_small_patches(mask, areaThr=100, structure=[[1, 1, 1], [1, 1, 1], [1, 1, 1]]):
    """
    remove small isolated patches
    """

    if mask.dtype == np.bool:
        pass
    elif issubclass(mask.dtype.type, np.integer):
        if np.amin(mask) < 0 or np.amax(mask) > 1: raise ValueError, 'Values of input image should be either 0 or 1.'
    else:
        raise TypeError, 'Data type of input image should be either np.bool or integer.'

    patches, n = ni.label(mask, structure)
    newMask = np.zeros(mask.shape, dtype=np.uint8)

    if n == 0:
        return newMask
    else:
        for i in range(1, n + 1):
            currPatch = np.zeros(mask.shape, dtype=np.uint8)
            currPatch[patches == i] = 1
            if np.sum(currPatch.flatten()) >= areaThr: newMask += currPatch

    return newMask.astype(np.bool)


def get_area_edges(img,
                   firstGaussianSigma=50.,
                   medianFilterWidth=100.,
                   areaThr=(0.1, 0.9),
                   edgeThrRange=(5, 16),
                   secondGaussianSigma=10.,
                   thr=0.2,
                   borderWidth=2,
                   lengthThr=20,
                   isPlot=True):
    """
    get binary edge of areas
    """

    img = img.astype(np.float)
    imgFlat = img - ni.filters.gaussian_filter(img, firstGaussianSigma)
    imgMedianFiltered = array_nor(ni.filters.median_filter(imgFlat, medianFilterWidth))
    imgPatch = np.array(imgMedianFiltered)
    imgPatch[imgMedianFiltered < areaThr[0]] = areaThr[0];
    imgPatch[imgMedianFiltered > areaThr[1]] = areaThr[1]
    imgPatch = (array_nor(imgPatch) * 255).astype(np.uint8)

    # plt.imshow(imgPatch,vmin=0,vmax=255,cmap='gray')
    # plt.show()

    #    import tifffile as tf
    #    tf.imsave('Rorb_example_vasMap_filtered.tif',array_nor(imgPatch.astype(np.float32)))


    cuttingStep = np.arange(edgeThrRange[0], edgeThrRange[1])
    cuttingStep = np.array([cuttingStep[0:-1], cuttingStep[1:]]).transpose()
    edge_cv2 = np.zeros(img.shape).astype(np.uint8)
    for i in range(cuttingStep.shape[0]):
        currEdge = cv2.Canny(imgPatch, cuttingStep[i, 0], cuttingStep[i, 1]) / 255
        edge_cv2 += currEdge
        if isPlot:
            if i == 1: firstEdgeSet = currEdge
            if i == cuttingStep.shape[0] - 1: lastEdgeSet = currEdge

    edgesF = ni.filters.gaussian_filter(edge_cv2.astype(np.float), secondGaussianSigma)

    edgesThr = np.zeros(edgesF.shape).astype(np.uint8)
    edgesThr[edgesF < thr] = 0;
    edgesThr[edgesF >= thr] = 1

    edgesThin = sm.skeletonize(edgesThr)

    edgesThin = remove_small_patches(edgesThin, lengthThr)
    if borderWidth > 1:
        edgesThick = ni.binary_dilation(edgesThin, iterations=borderWidth - 1)
    else:
        edgesThick = edgesThin

    if isPlot:
        displayEdges = np.zeros((edgesThick.shape[0], edgesThick.shape[1], 4)).astype(np.uint8)
        displayEdges[edgesThick == 1] = np.array([255, 0, 0, 255]).astype(np.uint8)
        displayEdges[edgesThick == 0] = np.array([0, 0, 0, 0]).astype(np.uint8)

        f, ax = plt.subplots(2, 5, figsize=(15, 5))
        ax[0, 0].imshow(img, cmap='gray');
        ax[0, 0].set_title('original image');
        ax[0, 0].axis('off')
        ax[0, 1].imshow(imgFlat, cmap='gray');
        ax[0, 1].set_title('flattened image');
        ax[0, 1].axis('off')
        ax[0, 2].imshow(imgPatch, cmap='gray');
        ax[0, 2].set_title('image for edge detection');
        ax[0, 2].axis('off')
        ax[0, 3].imshow(firstEdgeSet, cmap='gray');
        ax[0, 3].set_title('first edge set');
        ax[0, 3].axis('off')
        ax[0, 4].imshow(lastEdgeSet, cmap='gray');
        ax[0, 4].set_title('last edge set');
        ax[0, 4].axis('off')
        ax[1, 0].imshow(edgesF, cmap='hot');
        ax[1, 0].set_title('filtered edges sum');
        ax[1, 0].axis('off')
        ax[1, 1].imshow(edgesThr, cmap='gray');
        ax[1, 1].set_title('binary thresholded edges');
        ax[1, 1].axis('off')
        ax[1, 2].imshow(edgesThick, cmap='gray');
        ax[1, 2].set_title('binary edges');
        ax[1, 2].axis('off')
        ax[1, 3].imshow(img, cmap='gray');
        ax[1, 3].imshow(displayEdges);
        ax[1, 3].set_title('original image with edges');
        ax[1, 3].axis('off')
        ax[1, 4].imshow(imgPatch, cmap='gray');
        ax[1, 4].imshow(displayEdges);
        ax[1, 4].set_title('blurred image with edges');
        ax[1, 4].axis('off')
        plt.tight_layout()
        return edgesThick.astype(np.bool), f

    else:
        return edgesThick.astype(np.bool)


def z_downsample(img, downSampleRate):
    """
    downsample input image in z direction
    """

    if len(img.shape) != 3:
        raise ValueError, 'Input array shoud be 3D!'

    newFrameNum = (img.shape[0] - (img.shape[0] % downSampleRate)) / downSampleRate
    newImg = np.empty((newFrameNum, img.shape[1], img.shape[2]), dtype=img.dtype)

    print 'Start downsampling...'
    for i in range(newFrameNum):
        #            print (float(i)*100/newFrameNum),'%'
        currChunk = img[i * downSampleRate:(i + 1) * downSampleRate, :, :].astype(np.float)
        currFrame = np.mean(currChunk, axis=0)
        newImg[i, :, :] = currFrame.astype(img.dtype)
    print 'End of downsampling.'
    return newImg


def get_masks(labeled, minArea=None, maxArea=None, isSort=True, keyPrefix=None, labelLength=3):
    """
    get mask dictionary from labeled maps (labeled by scipy.ndimage.label function)
    area range of each mask was defined by minArea and maxArea
    isSort: if True, sort masks by areas, big to small
    keyPrefix: the prefix of key
    labelLength: the number of characters of key
    """

    maskNum = np.max(labeled.flatten())
    masks = {}
    for i in range(1, maskNum + 1):
        currMask = np.zeros(labeled.shape, dtype=np.uint8)
        currMask[labeled == i] = 1

        if minArea is not None and np.sum(currMask.flatten()) < minArea:
            pass
        elif maxArea is not None and np.sum(currMask.flatten()) > maxArea:
            pass
        else:
            if keyPrefix is not None:
                currKey = keyPrefix + '.' + ft.int2str(i, labelLength)
            else:
                currKey = ft.int2str(i, labelLength)
            masks.update({currKey: currMask})

    if isSort:
        masks = sort_masks(masks, keyPrefix=keyPrefix, labelLength=labelLength)

    return masks


def get_marked_masks(labeled, markCoor):
    """
    return one binary masks which contain the marked coordinate
    labeled (maps labeled by scipy.ndimage.label function)
    """

    masks = get_masks(labeled)
    for key, value in masks.iteritems():
        if hit_or_miss(markCoor, value): return value
    return None


def sort_masks(masks, keyPrefix='', labelLength=3):
    """
    sort a dictionary of binary masks, big to small
    """

    maskNum = len(masks.keys())
    order = []
    for key, mask in masks.iteritems():
        order.append([key, np.sum(mask.flatten())])

    order = sorted(order, key=lambda a: a[1], reverse=True)

    newMasks = {}
    for i in range(len(order)):
        if keyPrefix is not None:
            currKey = keyPrefix + '.' + ft.int2str(i, labelLength)
        else:
            currKey = ft.int2str(i, labelLength)
        newMasks.update({currKey: masks[order[i][0]]})
    return newMasks


def temp_downsample(A, rate, verbose=False):
    """
    down sample a 3-d array in 0 direction
    """

    if len(A.shape) != 3: raise ValueError, 'input array should be 3-d.'
    rate = int(rate)
    dataType = A.dtype
    newZDepth = (A.shape[0] - (A.shape[0] % rate)) / rate
    newA = np.empty((newZDepth, A.shape[1], A.shape[2]), dtype=dataType)

    for i in range(newZDepth):
        if verbose:
            print (float(i) * 100 / newZDepth), '%'
            currChunk = A[i * 3:(i + 1) * 3, :, :].astype(np.float)
            currFrame = np.mean(currChunk, axis=0)
            newA[i, :, :] = currFrame.astype(dataType)
    return newA


def get_average_movie(mov, frameTS, onsetTimes, chunkDur):
    """
    :param mov: image movie
    :param frameTS: the timestamps for each frame of the raw movie
    :param onsetTimes: time stamps of onset of each trigger
    :param chunkDur: duration of each chunk
    :return: averageed movie of all chunks
    """

    meanFrameDur = np.mean(np.diff(frameTS))

    chunkFrameDur = int(np.ceil(chunkDur / meanFrameDur))

    # print 'chunkDur:', chunkDur
    # print 'meanFrameDur:', meanFrameDur
    # print 'chunkFrameDur:', chunkFrameDur

    sumMov = None
    n = 0.

    for onset in onsetTimes:
        onsetFrameInd = np.argmin(np.abs(frameTS - onset))
        print 'Chunk:', int(
            n), '; Starting frame index:', onsetFrameInd, '; Ending frame index', onsetFrameInd + chunkFrameDur

        if onsetFrameInd + chunkFrameDur <= mov.shape[0]:
            if sumMov is None: sumMov = np.zeros((chunkFrameDur, mov.shape[1], mov.shape[2]))
            sumMov += mov[onsetFrameInd:onsetFrameInd + chunkFrameDur, :, :].astype(np.float32)
            n += 1.
        else:
            print 'Ending frame index (' + int(
                onsetFrameInd + chunkFrameDur) + ') is larger than frames in movie (' + int(
                mov.shape[0]) + '.\nExclude this trigger.'
            continue

    return sumMov.astype(np.float32) / n


class ROI(object):
    """
    class of binary ROI
    """

    def __init__(self, mask, pixelSize=None, pixelSizeUnit=None):
        """
        :param mask: 2-d array, if not binary, non-zero and non-nan pixel will be included in mask,
                     zero-pixel will be considered as background
        :param pixelSize: float, can be None, one value (square pixel) or (width, height) for non-square pixel
        :param pixelSizeUnit: str, the unit of pixel size
        """

        if len(mask.shape) != 2: raise ValueError, 'Input mask should be 2d.'

        self.dimension = mask.shape
        self.pixels = np.where(np.logical_and(mask != 0, ~np.isnan(mask)))

        if pixelSize is None:
            self.pixelSizeX = self.pixelSizeY = pixelSize
        elif (not hasattr(pixelSize, '__len__')):
            self.pixelSizeX = self.pixelSizeY = pixelSize
        elif len(pixelSize) == 2:
            self.pixelSizeY = pixelSize[0]; self.pixelSizeX = pixelSize[1]
        else:
            raise LookupError, 'pixel size should be either None or scalar or list(array) of two sclars!!'

        if pixelSize is None:
            self.pixelSizeUnit = None
        else:
            self.pixelSizeUnit = pixelSizeUnit

    def get_indices(self):
        """
        :return: index list of the pixels in the ROI
        """
        return self.pixels

    def get_binary_mask(self):
        """
        generate binary mask of the ROI, return 2d array, with 0s and 1s, dtype np.uint8
        """
        mask = np.zeros(self.dimension, dtype=np.uint8)
        mask[self.pixels] = 1
        return mask

    def get_nan_mask(self):
        """
        generate float mask of the ROI, return 2d array, with nans and 1s, dtype np.float32
        """
        mask = np.zeros(self.dimension, dtype=np.float32)
        mask[:] = np.nan
        mask[self.pixels] = 1
        return mask

    def get_pixel_area(self):
        """
        return the area coverage of the ROI
        """

        if (self.pixelSizeX is not None) and (self.pixelSizeX is not None):
            print 'returning area with unit:' + self.pixelSizeUnit + '^2'
            return float(len(self.pixels[0])) * self.pixelSizeX * self.pixelSizeY
        else:
            print 'returning area as pixel counts without unit.'
            return len(self.pixels[0])

    def get_center(self):
        """
        return the center coordinates [Y, X] of the centroid of the mask
        """
        return np.mean(np.array(self.pixels, dtype=np.float).transpose(), axis=0)

    def get_binary_trace(self, mov):
        """
        return trace of this ROI (binary format, 0s and 1s) in a given movie
        """
        binaryMask = self.get_binary_mask()
        trace = np.multiply(mov, np.array([binaryMask])).sum(axis=1).sum(axis=1)
        return trace

    def plot_binary_mask(self, plotAxis=None, color='#ff0000', alpha=1):
        """
        return display image (RGBA uint8 format) which can be plotted by plt.imshow, alpha: transparency 0-1
        """
        mask = self.get_binary_mask()
        displayImg = pt.binary_2_rgba(mask, foregroundColor=color, backgroundColor='#000000',
                                      foregroundAlpha=int(alpha * 255), backgroundAlpha=0)
        if plotAxis is None: f = plt.figure();plotAxis = f.add_subplot(111);plotAxis.imshow(displayImg,
                                                                                            interpolation='nearest')
        return displayImg

    def plot_binary_mask_border(self, **kwargs):
        pt.plot_mask_borders(self.get_nan_mask(), **kwargs)

    def to_h5_group(self, h5Group):
        """
        add attributes and dataset to a h5 data group
        """
        h5Group.attrs['dimension'] = self.dimension
        if self.pixelSizeX is None:
            h5Group.attrs['pixelSize'] = 'None'
        else:
            h5Group.attrs['pixelSize'] = [self.pixelSizeY, self.pixelSizeX]
        if self.pixelSizeUnit is None:
            h5Group.attrs['pixelSizeUnit'] = 'None'
        else:
            h5Group.attrs['pixelSizeUnit'] = self.pixelSizeUnit

        dataDict = dict(self.__dict__)
        _ = dataDict.pop('dimension');
        _ = dataDict.pop('pixelSizeX');
        _ = dataDict.pop('pixelSizeY');
        _ = dataDict.pop('pixelSizeUnit')
        for key, value in dataDict.iteritems():
            if value is None:
                h5Group.create_dataset(key, data='None')
            else:
                h5Group.create_dataset(key, data=value)

    @staticmethod
    def from_h5_group(h5Group):
        """
        load ROI (either ROI or WeightedROI) object from a hdf5 data group
        """

        dimension = h5Group.attrs['dimension']
        pixelSize = h5Group.attrs['pixelSize']
        if pixelSize == 'None': pixelSize = None
        pixelSizeUnit = h5Group.attrs['pixelSizeUnit']
        if pixelSizeUnit == 'None': pixelSizeUnit = None
        pixels = h5Group['pixels'].value

        if 'weights' in h5Group.keys():
            weights = h5Group['weights'].value
            mask = np.zeros(dimension, dtype=np.float32);
            mask[tuple(pixels)] = weights
            return WeightedROI(mask, pixelSize=pixelSize, pixelSizeUnit=pixelSizeUnit)
        else:
            mask = np.zeros(dimension, dtype=np.uint8);
            mask[tuple(pixels)] = 1
            return ROI(mask, pixelSize=pixelSize, pixelSizeUnit=pixelSizeUnit)


class WeightedROI(ROI):
    def __init__(self, mask, pixelSize=None, pixelSizeUnit=None):
        super(WeightedROI, self).__init__(mask, pixelSize=pixelSize, pixelSizeUnit=pixelSizeUnit)
        self.weights = mask[self.pixels]

    def get_weighted_mask(self):
        mask = np.zeros(self.dimension, dtype=np.float32)
        mask[self.pixels] = self.weights
        return mask

    def get_weighted_nan_mask(self):
        mask = np.zeros(self.dimension, dtype=np.float32)
        mask[:] = np.nan
        mask[self.pixels] = self.weights
        return mask

    def get_weighted_center(self):
        pixelCor = np.array(self.pixels, dtype=np.float)
        center = np.sum(np.multiply(pixelCor, np.array(self.weights)), axis=1) / np.sum(self.weights)
        return center

    def get_weighted_center_in_coordinate(self, yCor, xCor):
        """
        return weighted center of the ROI in the coordinate system defined by np.meshgrid(xCor, yCor)
        """
        weightMask = self.get_weighted_mask()
        xMap, yMap = np.meshgrid(xCor, yCor)
        xCenter = np.sum((xMap * weightMask).flatten()) / np.sum(weightMask.flatten())
        yCenter = np.sum((yMap * weightMask).flatten()) / np.sum(weightMask.flatten())
        return [yCenter, xCenter]

    def plot_weighted_mask(self, plotAxis=None, color='#ff0000'):
        """
        return display image (RGBA uint8 format) which can be plotted by plt.imshow
        """
        mask = self.get_weighted_mask()
        displayImg = pt.scalar_2_rgba(mask, color=color)
        if plotAxis is None: f = plt.figure(); plotAxis = f.add_subplot(111); plotAxis.imshow(displayImg,
                                                                                              interpolation='nearest')
        return displayImg

    def get_weighted_trace(self, mov):
        """
        return trace of this ROI in a given movie
        """
        weightedMask = self.get_weighted_mask()
        trace = np.multiply(mov, np.array([weightedMask])).sum(axis=1).sum(axis=1)
        return trace

    @staticmethod
    def from_h5_group(h5Group):
        """
        load WeightedROI (either ROI or WeightedROI) object from a hdf5 data group
        """

        dimension = h5Group.attrs['dimension']
        pixelSize = h5Group.attrs['pixelSize']
        if pixelSize == 'None': pixelSize = None
        pixelSizeUnit = h5Group.attrs['pixelSizeUnit']
        if pixelSizeUnit == 'None': pixelSizeUnit = None
        pixels = h5Group['pixels'].value
        weights = h5Group['weights'].value
        mask = np.zeros(dimension, dtype=np.float32);
        mask[tuple(pixels)] = weights
        return WeightedROI(mask, pixelSize=pixelSize, pixelSizeUnit=pixelSizeUnit)


if __name__ == '__main__':
    print 'for debug'



'''
2017-10-31 by Jun Zhuang
this module provides analysis tools to extract information about visual stimuli
saved in the log pkl files.
'''

import numpy as np
import tools.FileTools as ft

class DisplayLogAnalyzer(object):
    """
    class to take display_log (.pkl) file, check its integrity and extract stimuli and display
    organize into stim_dict dictionary, which is a intermediate step to put visual display
    information into nwb files.
    """

    def __init__(self, log_path):

        self.log_dict = ft.loadFile(log_path)

        if not self.log_dict['presentation']['is_by_index']:
            raise NotImplementedError('The visual stimuli display should be indexed.')

        self.check_integrity()

    def check_integrity(self):

        print(self.log_dict['presentation']['frame_stats'])

        if not self.log_dict['presentation']['keep_display']:
            raise ValueError('Stimulus presentation did not end normally.')

        total_frame1 = len(self.log_dict['presentation']['displayed_frames'])
        total_frame2 = len(self.log_dict['presentation']['frame_ts_start'])
        total_frame3 = len(self.log_dict['presentation']['frame_ts_end'])
        total_frame4 = len(self.log_dict['stimulation']['index_to_display'])
        if not total_frame1 == total_frame2 == total_frame3 == total_frame4:
            print('\nNumber of displayed frames: {}.'.format(total_frame1))
            print('\nNumber of frame start timestamps: {}.'.format(total_frame2))
            print('\nNumber of frame end timestamps: {}.'.format(total_frame3))
            print('\nNumber of frames to be displayed: {}.'.format(total_frame4))
            raise ValueError('Numbers of total frames do not agree with each other from various places.')

        if max(self.log_dict['stimulation']['index_to_display']) >= \
                len(self.log_dict['stimulation']['frames_unique']):
            raise ValueError('Display index beyond number of unique frames.')

        if self.log_dict['stimulation']['stim_name'] == 'CombinedStimuli':
            stimuli_sequence_out = [f[0] for f in self.log_dict['presentation']['displayed_frames']]
            stimuli_sequence_out = list(set(stimuli_sequence_out))
            stimuli_sequence_out.sort()
            stimuli_sequence_in = self.log_dict['stimulation']['individual_logs'].keys()
            stimuli_sequence_in.sort()
            if stimuli_sequence_out != stimuli_sequence_in:
                raise ValueError('Output stimuli sequence does not match input stimuli sequence.')

    @property
    def num_frame_tot(self):
        return len(self.log_dict['presentation']['displayed_frames'])

    def get_stim_dict(self):
        """
        Returns
        -------
        stim_dict: dictionary
            the structure of this dictionary should look like this:

            {
             '000_UniformContrastRetinotopicMapping': {
                                                       ...
                                                       'stim_name' : '000_UniformContrastRetinotopicMapping',
                                                       'index_to_display': <index referencing 'frames_unique' field>
                                                       'timestamps': <index referencing entire display sequence,
                                                                     should match hardware vsync signal>
                                                       'frames_unique': list of tuple representing unique frames
                                                       ...
                                                       },
             '001_StimulusSeparatorRetinotopicMapping: {
                                                        ...
                                                        'stim_name' : '000_UniformContrastRetinotopicMapping',
                                                        'index_to_display': <index referencing 'frames_unique' field>
                                                        'timestamps': <index referencing entire display sequence,
                                                                       should match hardware vsync signal>
                                                        'frames_unique': list of tuple representing unique frames
                                                        ...
                                                        },
             ...
             }
        """

        comments = ''
        description = ''
        source = 'retinotopic_mapping package'
        stim_dict = {}

        # if multiple stimuli were displayed in a sequence
        if self.log_dict['stimulation']['stim_name'] == 'CombinedStimuli':
            curr_frame_ind = 0
            stim_ids = self.log_dict['stimulation']['individual_logs'].keys()
            stim_ids.sort()
            for stim_id in stim_ids:
                curr_dict = self.log_dict['stimulation']['individual_logs'][stim_id]
                curr_stim_name = stim_id + 'RetinotopicMapping'
                curr_dict['stim_name'] = curr_stim_name
                curr_num_frames = len(curr_dict['index_to_display'])
                curr_dict.update({'timestamps': np.arange(curr_num_frames, dtype=np.uint64) + curr_frame_ind,
                                  'comments': comments,
                                  'source': source,
                                  'description': description})
                curr_frame_ind = curr_frame_ind + curr_num_frames
                stim_dict.update({curr_stim_name: curr_dict})

        # if only one stimulus was displayed
        else:
            stim_name = self.log_dict['stimulation']['stim_name']
            if stim_name in ['UniformContrast', 'FlashingCircle', 'SparseNoise', 'LocallySparseNoise',
                             'DriftingGratingCirlce', 'StaticGratingCircle', 'StaticImages', 'StimulusSeparator']:
                curr_stim_name = '{:03d}_{}RetinotopicMapping'.format(0, stim_name)
                curr_dict = self.log_dict['stimulation']
                curr_dict['stim_name'] = curr_stim_name
                curr_dict.update({'timestamps': np.arange(self.num_frame_tot, dtype=np.uint64)})
            else:
                raise NotImplementedError('Do not understand stimulus: {}.'.format(stim_name))

            curr_dict.update({'comments': comments,
                              'source': source,
                              'description': description})
            stim_dict.update({curr_stim_name : curr_dict})

        return stim_dict







import numpy as np


def weighted_average(values, weights):
    if values.shape != weights.shape:
        raise ValueError('"values" and "weights" should have same shape.')

    if np.amin(weights) < 0.:
        raise ValueError('all weights should be no less than 0.')

    return np.sum((values * weights).flat) / np.sum(weights.flat)



# -*- coding: utf-8 -*-
"""
Notes
-----

"""
import numpy as np
import matplotlib.pyplot as plt
import retinotopic_mapping.tools.ImageAnalysis as ia


class Monitor(object):
    """
    monitor object created by Jun, has the method "remap" to generate the
    spherical corrected coordinates in degrees

    This object contains the relevant data for the monitor used within a
    given experimental setup. When initialized, the rectangular coordinates
    of the pixels on the monitor are computed and stored as `lin_coord_x`,
    `lin_coord_y`. The rectangular coordinates are then transformed and
    warped by calling the `remap` method to populate the `deg_coord_x` and
    `deg_coord_y` attributes.

    Parameters
    ----------
    resolution : tuple
        value of the monitor resolution
    dis : float
         distance from eyeball to monitor (in cm)
    mon_width_cm : float
        width of monitor (in cm)
    mon_height_cm : float
        height of monitor (in cm)
    C2T_cm : float
        distance from gaze center to monitor top
    C2A_cm : float
        distance from gaze center to anterior edge of the monitor
    center_coordinates : tuple of two floats
        (altitude, azimuth), in degrees. the coordinates of the projecting point
        from the eye ball to the monitor. This allows to place the display monitor
        in any arbitrary position.
    visual_field : str from {'right','left'}, optional
        the eye that is facing the monitor, defaults to 'right'
    deg_coord_x : ndarray, optional
         array of warped x pixel coordinates, defaults to `None`
    deg_coord_y : ndarray, optional
         array of warped y pixel coordinates, defaults to `None`
    name : str, optional
         name of the monitor, defaults to `testMonitor`
    gamma : optional
         for gamma correction, defaults to `None`
    gamma_grid : optional
         for gamme correction, defaults to `None`
    luminance : optional
         monitor luminance, defaults to `None`
    downsample_rate : int, optional
         downsample rate of monitor pixels, defaults to 10
    refresh_rate : float, optional
        the refresh rate of the monitor in Hz, defaults to 60
    """

    def __init__(self,
                 resolution,
                 dis,
                 mon_width_cm,
                 mon_height_cm,
                 C2T_cm,
                 C2A_cm,
                 center_coordinates=(0., 60.),
                 visual_field='right',
                 deg_coord_x=None,
                 deg_coord_y=None,
                 name='testMonitor',
                 gamma=None,
                 gamma_grid=None,
                 luminance=None,
                 downsample_rate=10,
                 refresh_rate=60.):
        """
        Initialize monitor object.

        """

        if resolution[0] % downsample_rate != 0 \
                or resolution[1] % downsample_rate != 0:
            raise ArithmeticError('Resolution pixel numbers are not '
                                  'divisible by down sampling rate.')

        self.resolution = resolution
        self.dis = dis
        self.mon_width_cm = mon_width_cm
        self.mon_height_cm = mon_height_cm
        self.C2T_cm = C2T_cm
        self.C2A_cm = C2A_cm
        self.center_coordinates = center_coordinates
        self.visual_field = visual_field
        self.deg_coord_x = deg_coord_x
        self.deg_coord_y = deg_coord_y
        self.name = name
        self.downsample_rate = downsample_rate
        self.gamma = gamma
        self.gamma_grid = gamma_grid
        self.luminance = luminance
        self.refresh_rate = 60

        # distance form projection point of the eye to bottom of the monitor
        self.C2B_cm = self.mon_height_cm - self.C2T_cm
        # distance form projection point of the eye to right of the monitor
        self.C2P_cm = self.mon_width_cm - self.C2A_cm

        resolution = [0, 0]
        resolution[0] = self.resolution[0] / downsample_rate
        resolution[1] = self.resolution[1] / downsample_rate

        map_coord_x, map_coord_y = np.meshgrid(range(resolution[1]),
                                               range(resolution[0]))

        if self.visual_field == "left":
            map_x = np.linspace(self.C2A_cm, -1.0 * self.C2P_cm, resolution[1])

        if self.visual_field == "right":
            map_x = np.linspace(-1 * self.C2A_cm, self.C2P_cm, resolution[1])

        map_y = np.linspace(self.C2T_cm, -1.0 * self.C2B_cm, resolution[0])
        old_map_x, old_map_y = np.meshgrid(map_x, map_y, sparse=False)

        self.lin_coord_x = old_map_x
        self.lin_coord_y = old_map_y

        self.remap()

    def set_gamma(self, gamma, gamma_grid):
        self.gamma = gamma
        self.gamma_grid = gamma_grid

    def set_luminance(self, luminance):
        self.luminance = luminance

    def set_downsample_rate(self, downsample_rate):

        if self.resolution[0] % downsample_rate != 0 \
                or self.resolution[1] % downsample_rate != 0:
            raise ArithmeticError('Resolution pixel numbers are not divisible by down sampling rate.')

        self.downsample_rate = downsample_rate

        resolution = [0, 0]
        resolution[0] = self.resolution[0] / downsample_rate
        resolution[1] = self.resolution[1] / downsample_rate

        # map_coord_x, map_coord_y = np.meshgrid(range(resolution[1]),
        #                                        range(resolution[0]))

        if self.visual_field == "left":
            map_x = np.linspace(self.C2A_cm, -1.0 * self.C2P_cm, resolution[1])

        if self.visual_field == "right":
            map_x = np.linspace(-1 * self.C2P_cm, self.C2P_cm, resolution[1])

        map_y = np.linspace(self.C2T_cm, -1.0 * self.C2B_cm, resolution[0])
        old_map_x, old_map_y = np.meshgrid(map_x, map_y, sparse=False)

        self.lin_coord_x = old_map_x
        self.lin_coord_y = old_map_y

        self.remap()

    def remap(self):
        """
        warp the linear pixel coordinates to a spherical corrected representation.

        Function is called when the monitor object is initialized and populate
        the `deg_coord_x` and `deg_coord_y` attributes.
        """

        resolution = [0, 0]
        resolution[0] = self.resolution[0] / self.downsample_rate
        resolution[1] = self.resolution[1] / self.downsample_rate

        map_coord_x, map_coord_y = np.meshgrid(range(resolution[1]),
                                               range(resolution[0]))

        new_map_x = np.zeros(resolution, dtype=np.float32)
        new_map_y = np.zeros(resolution, dtype=np.float32)

        for j in range(resolution[1]):
            new_map_x[:, j] = ((180.0 / np.pi) *
                               np.arctan(self.lin_coord_x[0, j] / self.dis))
            dis2 = np.sqrt(np.square(self.dis) +
                           np.square(self.lin_coord_x[0, j]))

            for i in range(resolution[0]):
                new_map_y[i, j] = ((180.0 / np.pi) *
                                   np.arctan(self.lin_coord_y[i, 0] / dis2))

        self.deg_coord_x = new_map_x + self.center_coordinates[1]
        self.deg_coord_y = new_map_y + self.center_coordinates[0]

    def plot_map(self):

        resolution = [0, 0]
        resolution[0] = self.resolution[0] / self.downsample_rate
        resolution[1] = self.resolution[1] / self.downsample_rate

        mapcorX, mapcorY = np.meshgrid(range(resolution[1]), range(resolution[0]))

        f1 = plt.figure(figsize=(12, 7))
        f1.suptitle('Remap monitor', fontsize=14, fontweight='bold')

        OMX = plt.subplot(221)
        OMX.set_title('Linear Map X (cm)')
        currfig = plt.imshow(self.lin_coord_x)
        levels1 = range(int(np.floor(self.lin_coord_x.min() / 10) * 10),
                        int((np.ceil(self.lin_coord_x.max() / 10) + 1) * 10), 10)
        im1 = plt.contour(mapcorX, mapcorY, self.lin_coord_x, levels1, colors='k', linewidth=2)
        #        plt.clabel(im1, levels1, fontsize = 10, inline = 1, fmt='%2.1f')
        f1.colorbar(currfig, ticks=levels1)
        plt.gca().set_axis_off()

        OMY = plt.subplot(222)
        OMY.set_title('Linear Map Y (cm)')
        currfig = plt.imshow(self.lin_coord_y)
        levels2 = range(int(np.floor(self.lin_coord_y.min() / 10) * 10),
                        int((np.ceil(self.lin_coord_y.max() / 10) + 1) * 10), 10)
        im2 = plt.contour(mapcorX, mapcorY, self.lin_coord_y, levels2, colors='k', linewidth=2)
        #        plt.clabel(im2, levels2, fontsize = 10, inline = 1, fmt='%2.2f')
        f1.colorbar(currfig, ticks=levels2)
        plt.gca().set_axis_off()

        NMX = plt.subplot(223)
        NMX.set_title('Spherical Map X (deg)')
        currfig = plt.imshow(self.deg_coord_x)
        levels3 = range(int(np.floor(self.deg_coord_x.min() / 10) * 10),
                        int((np.ceil(self.deg_coord_x.max() / 10) + 1) * 10), 10)
        im3 = plt.contour(mapcorX, mapcorY, self.deg_coord_x, levels3, colors='k', linewidth=2)
        #        plt.clabel(im3, levels3, fontsize = 10, inline = 1, fmt='%2.1f')
        f1.colorbar(currfig, ticks=levels3)
        plt.gca().set_axis_off()
        #
        NMY = plt.subplot(224)
        NMY.set_title('Spherical Map Y (deg)')
        currfig = plt.imshow(self.deg_coord_y)
        levels4 = range(int(np.floor(self.deg_coord_y.min() / 10) * 10),
                        int((np.ceil(self.deg_coord_y.max() / 10) + 1) * 10), 10)
        im4 = plt.contour(mapcorX, mapcorY, self.deg_coord_y, levels4, colors='k', linewidth=2)
        #        plt.clabel(im4, levels4, fontsize = 10, inline = 1, fmt='%2.1f')
        f1.colorbar(currfig, ticks=levels4)
        plt.gca().set_axis_off()

    def generate_lookup_table(self):
        """
        generate lookup talbe between degree corrdinates and linear corrdinates
        return two matrix:
        lookupI: i index in linear matrix to this pixel after warping
        lookupJ: j index in linear matrix to this pixel after warping
        """

        # length of one degree on monitor at gaze point
        degDis = np.tan(np.pi / 180) * self.dis

        # generate degree coordinate without warpping
        degNoWarpCorX = self.lin_coord_x / degDis
        degNoWarpCorY = self.lin_coord_y / degDis

        # deg coordinates
        degCorX = self.deg_coord_x + self.center_coordinates[0]
        degCorY = self.deg_coord_y + self.center_coordinates[1]

        lookupI = np.zeros(degCorX.shape).astype(np.int32)
        lookupJ = np.zeros(degCorX.shape).astype(np.int32)

        for j in xrange(lookupI.shape[1]):
            currDegX = degCorX[0, j]
            diffDegX = degNoWarpCorX[0, :] - currDegX
            IndJ = np.argmin(np.abs(diffDegX))
            lookupJ[:, j] = IndJ

            for i in xrange(lookupI.shape[0]):
                currDegY = degCorY[i, j]
                diffDegY = degNoWarpCorY[:, IndJ] - currDegY
                indI = np.argmin(np.abs(diffDegY))
                lookupI[i, j] = indI

        return lookupI, lookupJ

    def warp_images(self, imgs, center_coor, deg_per_pixel=0.1, is_luminance_correction=True):
        """
        warp a image stack into visual degree coordinate system

        parameters
        ----------
        imgs : ndarray
            should be 2d or 3d, if 3d, axis will be considered as frame x rows x width
        center_coor : list or tuple of two floats
            the visual degree coordinates of the center of the image (altitude, azimuth)
        deg_per_pixel : float or list/tuple of two floats
            size of original pixel in visual degrees, (altitude, azimuth), if float, assume
            sizes in both dimension are the same
        is_luminance_correction : bool
            if True, wrapped images will have mean intensity equal 0, and values will be
            scaled up to reach minimum equal -1. or maximum equal 1.

        returns
        -------
        imgs_wrapped : 3d array, np.float32
            wrapped images, each frame should have exact same size of down sampled monitor
            resolution. the region on the monitor not covered by the image will have value
            of np.nan. value range [-1., 1.]
        coord_alt_wrapped : 2d array, np.float32
            the altitude coordinates of all pixels in the wrapped images in visual degrees.
            should have the same shape as each frame in 'imgs_wrapped'.
        coord_azi_wrapped : 2d array, np.float32
            the azimuth coordinates of all pixels in the wrapped images in visual degrees.
            should have the same shape as each frame in 'imgs_wrapped'.
        imgs_dewrapped : 3d array, dtype same as imgs
            unwrapped images, same dimension as input image stack. the region of original
            image that was not got displayed (outside of the monitor) will have value of
            np.nan. value range [-1., 1.]
        coord_alt_dewrapped : 2d array, np.float32
            the altitude coordinates of all pixels in the dewrapped images in visual degrees.
            should have the same shape as each frame in 'imgs_dewrapped'.
        coord_azi_dewrapped : 2d array, np.float32
            the azimuth coordinates of all pixels in the dewrapped images in visual degrees.
            should have the same shape as each frame in 'imgs_dewrapped'.
        """

        try:
            deg_per_pixel_alt = abs(float(deg_per_pixel[0]))
            deg_per_pixel_azi = abs(float(deg_per_pixel[1]))
        except TypeError:
            deg_per_pixel_alt = deg_per_pixel_azi = deg_per_pixel

        if len(imgs.shape) == 2:
            imgs_raw = np.array([imgs])
        elif len(imgs.shape) == 3:
            imgs_raw = imgs
        else:
            raise ValueError('input "imgs" should be 2d or 3d array.')

        # generate raw image pixel coordinates in visual degrees
        alt_start = center_coor[0] + (imgs_raw.shape[1] / 2) * deg_per_pixel_alt
        alt_axis = alt_start - np.arange(imgs_raw.shape[1]) * deg_per_pixel_alt
        azi_start = center_coor[1] - (imgs_raw.shape[2] / 2) * deg_per_pixel_azi
        azi_axis = np.arange(imgs_raw.shape[2]) * deg_per_pixel_azi + azi_start
        # img_coord_azi, img_coord_alt = np.meshgrid(azi_axis, alt_axis)

        # initialize output array
        imgs_wrapped = np.zeros((imgs_raw.shape[0],
                                 self.deg_coord_x.shape[0],
                                 self.deg_coord_x.shape[1]), dtype=np.float32)
        imgs_wrapped[:] = np.nan

        # for cropping imgs_raw
        x_min = x_max = y_max = y_min = None

        # for testing
        # img_count = np.zeros((imgs_raw.shape[1], imgs_raw.shape[2]), dtype=np.uint32)

        # loop through every display (wrapped) pixel
        for ii in range(self.deg_coord_x.shape[0]):
            for jj in range(self.deg_coord_x.shape[1]):

                # the wrapped coordinate of current display pixel [alt, azi]
                coord_w = [self.deg_coord_y[ii, jj], self.deg_coord_x[ii, jj]]

                # if the wrapped coordinates of current display pixel is covered
                # by the raw image
                if alt_axis[0] >= coord_w[0] >= alt_axis[-1] and \
                                        azi_axis[0] <= coord_w[1] <= azi_axis[-1]:

                    # get raw pixels arround the wrapped coordinates of current display pixel
                    u = (alt_axis[0] - coord_w[0]) / deg_per_pixel_alt
                    l = (coord_w[1] - azi_axis[0]) / deg_per_pixel_azi

                    # for testing:
                    # img_count[int(u), int(l)] += 1

                    if (u == round(u) and l == round(l)):  # right hit on one raw pixel
                        imgs_wrapped[:, ii, jj] = imgs_raw[:, int(u), int(l)]

                        # for cropping
                        if x_min is None:
                            x_min = x_max = l
                            y_min = y_max = u
                        else:
                            x_min = min(x_min, l)
                            x_max = max(x_max, l)
                            y_min = min(y_min, u)
                            y_max = max(y_max, u)

                    else:
                        u = int(u)
                        b = u + 1
                        l = int(l)
                        r = l + 1
                        w_ul = 1. / ia.distance(coord_w, [alt_axis[u], azi_axis[l]])
                        w_bl = 1. / ia.distance(coord_w, [alt_axis[b], azi_axis[l]])
                        w_ur = 1. / ia.distance(coord_w, [alt_axis[u], azi_axis[r]])
                        w_br = 1. / ia.distance(coord_w, [alt_axis[b], azi_axis[r]])

                        w_sum = w_ul + w_bl + w_ur + w_br

                        imgs_wrapped[:, ii, jj] = (imgs_raw[:, u, l] * w_ul +
                                                   imgs_raw[:, b, l] * w_bl +
                                                   imgs_raw[:, u, r] * w_ur +
                                                   imgs_raw[:, b, r] * w_br) / w_sum

                        # for cropping
                        if x_min is None:
                            x_min = l
                            x_max = l + 1
                            y_min = u
                            y_max = u + 1
                        else:
                            x_min = min(x_min, l)
                            x_max = max(x_max, l + 1)
                            y_min = min(y_min, u)
                            y_max = max(y_max, u + 1)

        # for testing
        # plt.imshow(img_count, interpolation='bicubic')
        # plt.colorbar()
        # plt.show()

        if is_luminance_correction:
            for frame_ind in range(imgs_wrapped.shape[0]):
                curr_frame = imgs_wrapped[frame_ind]
                curr_mean = np.nanmean(curr_frame.flat)
                curr_frame = curr_frame - curr_mean
                curr_amp = np.max([np.nanmax(curr_frame.flat), abs(np.nanmin(curr_frame.flat))])
                curr_frame = curr_frame / curr_amp
                imgs_wrapped[frame_ind] = curr_frame

        # crop image
        alt_range = np.logical_and(np.arange(imgs_raw.shape[1]) >= y_min,
                                   np.arange(imgs_raw.shape[1]) <= y_max)
        azi_range = np.logical_and(np.arange(imgs_raw.shape[2]) >= x_min,
                                   np.arange(imgs_raw.shape[2]) <= x_max)

        # print imgs_raw.shape
        # print imgs_raw.shape
        # print alt_range.shape
        # print azi_range.shape
        # print np.sum(alt_range)
        # print np.sum(azi_range)

        imgs_dewrapped = imgs_raw[:, alt_range, :]
        imgs_dewrapped = imgs_dewrapped[:, :, azi_range]

        # get degree coordinats of dewrapped images
        deg_coord_alt_ax_dewrapped = alt_axis[alt_range]
        deg_coord_azi_ax_dewrapped = azi_axis[azi_range]
        deg_coord_azi_dewrapped, deg_coord_alt_dewrapped = np.meshgrid(deg_coord_azi_ax_dewrapped,
                                                                       deg_coord_alt_ax_dewrapped)
        deg_coord_alt_dewrapped = deg_coord_alt_dewrapped.astype(np.float32)
        deg_coord_azi_dewrapped = deg_coord_azi_dewrapped.astype(np.float32)

        return imgs_wrapped, self.deg_coord_y, self.deg_coord_x, imgs_dewrapped, deg_coord_alt_dewrapped, \
               deg_coord_azi_dewrapped


class Indicator(object):
    """
    flashing indicator for photodiode

    Parameters
    ----------
    monitor : monitor object
        The monitor used within the experimental setup
    width_cm : float, optional
        width of the size of the indicator in cm, defaults to `3.`
    height_cm : float, optional
         height of the size of the indicator in cm, defaults to `3.`
    position : str from {'northeast','northwest','southwest','southeast'}
         the placement of the indicator, defaults to 'northeast'
    is_sync : bool, optional
         determines whether the indicator is synchronized with the stimulus,
         defaults to True.
    freq : float, optional
        frequency of photodiode, defaults to `2.`
    """

    def __init__(self,
                 monitor,
                 width_cm=3.,
                 height_cm=3.,
                 position='northeast',
                 is_sync=True,
                 freq=2.):
        """
        Initialize indicator object
        """

        self.monitor = monitor
        self.width_cm = width_cm
        self.height_cm = height_cm
        self.width_pixel, self.height_pixel = self.get_size_pixel()
        self.position = position
        self.center_width_pixel, self.center_height_pixel = self.get_center()
        self.is_sync = is_sync

        if is_sync == False:
            self.freq = freq  # if not synchronized with stimulation, self update frquency of the indicator
            self.frame_num = self.get_frames()
        else:
            self.freq = None
            self.frame_num = None

    def get_size_pixel(self):

        screen_width = (self.monitor.resolution[1] /
                        self.monitor.downsample_rate)
        screen_height = (self.monitor.resolution[0] /
                         self.monitor.downsample_rate)

        indicator_width = int((self.width_cm / self.monitor.mon_width_cm) *
                              screen_width)
        indicator_height = int((self.height_cm / self.monitor.mon_height_cm) *
                               screen_height)

        return indicator_width, indicator_height

    def get_center(self):

        screen_width = (self.monitor.resolution[1] /
                        self.monitor.downsample_rate)
        screen_height = (self.monitor.resolution[0] /
                         self.monitor.downsample_rate)

        if self.position == 'northeast':
            center_width = screen_width - self.width_pixel / 2
            center_height = self.height_pixel / 2

        elif self.position == 'northwest':
            center_width = self.width_pixel / 2
            center_height = self.height_pixel / 2

        elif self.position == 'southeast':
            center_width = screen_width - self.width_pixel / 2
            center_height = screen_height - self.height_pixel / 2

        elif self.position == 'southwest':
            center_width = self.width_pixel / 2
            center_height = screen_height - self.height_pixel / 2
        else:
            raise LookupError('`position` attribute not in '
                              '{"northeast","northwest","southeast","southwest"}.')

        return int(center_width), int(center_height)

    def get_frames(self):
        """
        if not synchronized with stimulation, get frame numbers of each update
        of indicator
        """

        refresh_rate = self.monitor.refresh_rate

        if refresh_rate % self.freq != 0:
            raise ArithmeticError("`freq` not divisble by monitor ref rate.")

        return refresh_rate / self.freq



# -*- coding: utf-8 -*-
"""
Created on Fri Oct 31 11:07:20 2014

@author: junz
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
import matplotlib.colors as col
import scipy.ndimage as ni
import tifffile as tf
import ImageAnalysis as ia
import cv2


def get_rgb(colorStr):
    """
    get R,G,B int value from a hex color string
    """
    return int(colorStr[1:3], 16), int(colorStr[3:5], 16), int(colorStr[5:7], 16)


def get_color_str(R, G, B):
    """
    get hex color string from R,G,B value (integer with uint8 format)
    """
    if not (isinstance(R, (int, long)) and isinstance(G, (int, long)) and isinstance(G, (int, long))):
        raise TypeError, 'Input R, G and B should be integer!'

    if not ((0 <= R <= 255) and (0 <= G <= 255) and (
            0 <= B <= 255)): raise ValueError, 'Input R, G and B should between 0 and 255!'
    return '#' + ''.join(map(chr, (R, G, B))).encode('hex')


def binary_2_rgba(img, foregroundColor='#ff0000', backgroundColor='#000000', foregroundAlpha=255, backgroundAlpha=0):
    """
    generate display image in (RGBA).(np.uint8) format which can be displayed by imshow
    :param img: input image, should be a binary array (np.bool, or np.(u)int
    :param foregroundColor: color for 1 in the array, RGB str, i.e. '#ff0000'
    :param backgroundColor: color for 0 in the array, RGB str, i.e. '#ff00ff'
    :param foregroundAlpha: alpha for 1 in the array, int, 0-255
    :param backgroundAlpha: alpha for 1 in the array, int, 0-255
    :return: displayImg, (RGBA).(np.uint8) format, ready for imshow
    """

    if img.dtype == np.bool:
        pass
    elif issubclass(img.dtype.type, np.integer):
        if np.amin(img) < 0 or np.amax(img) > 1: raise ValueError, 'Values of input image should be either 0 or 1.'
    else:
        raise TypeError, 'Data type of input image should be either np.bool or integer.'

    if type(foregroundAlpha) is int:
        if foregroundAlpha < 0 or foregroundAlpha > 255: raise ValueError, 'Value of foreGroundAlpha should be between 0 and 255.'
    else:
        raise TypeError, 'Data type of foreGroundAlpha should be integer.'

    if type(backgroundAlpha) is int:
        if backgroundAlpha < 0 or backgroundAlpha > 255: raise ValueError, 'Value of backGroundAlpha should be between 0 and 255.'
    else:
        raise TypeError, 'Data type of backGroundAlpha should be integer.'

    fR, fG, fB = get_rgb(foregroundColor)
    bR, bG, bB = get_rgb(backgroundColor)

    displayImg = np.zeros((img.shape[0], img.shape[1], 4)).astype(np.uint8)
    displayImg[img == 1] = np.array([fR, fG, fB, foregroundAlpha]).astype(np.uint8)
    displayImg[img == 0] = np.array([bR, bG, bB, backgroundAlpha]).astype(np.uint8)

    return displayImg


def scalar_2_rgba(img, color='#ff0000'):
    """
    generate display a image in (RGBA).(np.uint8) format which can be displayed by imshow
    alpha is defined by values in the img
    :param img: input image
    :param alphaMatrix: matrix of alpha
    :param foreGroundColor: color for 1 in the array, RGB str, i.e. '#ff0000'
    :return: displayImg, (RGBA).(np.uint8) format, ready for imshow
    """

    R, G, B = get_rgb(color)

    RMatrix = (R * ia.array_nor(img.astype(np.float32))).astype(np.uint8)
    GMatrix = (G * ia.array_nor(img.astype(np.float32))).astype(np.uint8)
    BMatrix = (B * ia.array_nor(img.astype(np.float32))).astype(np.uint8)

    alphaMatrix = (ia.array_nor(img.astype(np.float32)) * 255).astype(np.uint8)

    displayImg = np.zeros((img.shape[0], img.shape[1], 4)).astype(np.uint8)
    displayImg[:, :, 0] = RMatrix;
    displayImg[:, :, 1] = GMatrix;
    displayImg[:, :, 2] = BMatrix;
    displayImg[:, :, 3] = alphaMatrix

    return displayImg


def bar_graph(left,
              height,
              error,
              errorDir='both',  # 'both', 'positive' or 'negative'
              width=0.1,
              plotAxis=None,
              lw=3,
              faceColor='#000000',
              edgeColor='none',
              capSize=10,
              label=None
              ):
    """
    plot a single bar with error bar
    """

    if not plotAxis:
        f = plt.figure()
        plotAxis = f.add_subplot(111)

    if errorDir == 'both':
        yerr = error
    elif errorDir == 'positive':
        yerr = [[0], [error]]
    elif errorDir == 'negative':
        yerr = [[error], [0]]

    plotAxis.errorbar(left + width / 2,
                      height,
                      yerr=yerr,
                      lw=lw,
                      capsize=capSize,
                      capthick=lw,
                      color=edgeColor)

    plotAxis.bar(left,
                 height,
                 width=width,
                 color=faceColor,
                 edgecolor=edgeColor,
                 lw=lw,
                 label=label)

    return plotAxis


def random_color(numOfColor=10):
    """
    generate as list of random colors
    """
    numOfColor = int(numOfColor)

    colors = []

    Cmatrix = (np.random.rand(numOfColor, 3) * 255).astype(np.uint8)

    for i in range(numOfColor):

        r = hex(Cmatrix[i][0]).split('x')[1]
        if len(r) == 1:
            r = '0' + r

        g = hex(Cmatrix[i][1]).split('x')[1]
        if len(g) == 1:
            g = '0' + g

        b = hex(Cmatrix[i][2]).split('x')[1]
        if len(b) == 1:
            b = '0' + b

        colors.append('#' + r + g + b)

    return colors


def show_movie(path,  # tif file path or numpy arrary of the movie
               mode='raw',  # 'raw', 'dF' or 'dFoverF'
               baselinePic=None,  # picuture of baseline
               baselineType='mean',  # way to calculate baseline
               cmap='gray'):
    """
    plot tf movie in the way defined by mode
    """

    if isinstance(path, str):
        rawMov = tf.imread(path)
    elif isinstance(path, np.ndarray):
        rawMov = path

    if mode == 'raw':
        mov = rawMov
    else:
        _, dFMov, dFoverFMov = ia.normalize_movie(rawMov,
                                                  baselinePic=baselinePic,
                                                  baselineType=baselineType)
        if mode == 'dF':
            mov = dFMov
        elif mode == 'dFoverF':
            mov = dFoverFMov
        else:
            raise LookupError, 'The "mode" should be "raw", "dF" or "dFoverF"!'

    if isinstance(path, str):
        tf.imshow(mov,
                  cmap=cmap,
                  vmax=np.amax(mov),
                  vmin=np.amin(mov),
                  title=mode + ' movie of ' + path)
    elif isinstance(path, np.ndarray):
        tf.imshow(mov,
                  cmap=cmap,
                  vmax=np.amax(mov),
                  vmin=np.amin(mov),
                  title=mode + ' Movie')

    return mov


def standalone_color_bar(vmin, vmax, cmap, sectionNum=10):
    """
    plot a stand alone color bar.
    """

    a = np.array([[vmin, vmax]])

    plt.figure(figsize=(0.1, 9))

    img = plt.imshow(a, cmap=cmap, vmin=vmin, vmax=vmax)
    plt.gca().set_visible(False)
    cbar = plt.colorbar()
    cbar.set_ticks(np.linspace(vmin, vmax, num=sectionNum + 1))


def alpha_blending(image, alphaData, vmin, vmax, cmap='Paired', sectionNum=10, background=-1, interpolation='nearest',
                   isSave=False, savePath=None):
    """
    Generate image with transparency weighted by another matrix.

    Plot numpy array 'image' with colormap 'cmap'. And define the tranparency
    of each pixel by the value in another numpy array alphaData.

    All the elements in alphaData should be non-negative.
    """

    if image.shape != alphaData.shape:
        raise LookupError, '"image" and "alphaData" should have same shape!!'

    if np.amin(alphaData) < 0:
        raise ValueError, 'All the elements in alphaData should be bigger than zero.'

    # normalize image
    image[image > vmax] = vmax
    image[image < vmin] = vmin

    image = (image - vmin) / (vmax - vmin)

    # get colored image of image
    exec ('colorImage = cm.' + cmap + '(image)')

    # normalize alphadata
    alphaDataNor = alphaData / np.amax(alphaData)
    alphaDataNor = np.sqrt(alphaDataNor)

    colorImage[:, :, 3] = alphaDataNor

    # plt.figure()
    # plot dummy figure for colorbar
    a = np.array([[vmin, vmax]])
    plt.imshow(a, cmap=cmap, vmin=vmin, vmax=vmax, alpha=0)
    # plt.gca().set_visible(False)
    cbar = plt.colorbar()
    cbar.set_ticks(np.linspace(vmin, vmax, num=sectionNum + 1))
    cbar.set_alpha(1)
    cbar.draw_all()

    # generate black background
    b = np.array(colorImage)
    b[:] = background
    b[:, :, 3] = 1
    plt.imshow(b, cmap='gray')

    # plot map
    plt.imshow(colorImage, interpolation=interpolation)

    return colorImage


def plot_mask(mask, plotAxis=None, color='#ff0000', zoom=1, borderWidth=None, closingIteration=None):
    """
    plot mask borders in a given color
    """

    if not plotAxis:
        f = plt.figure()
        plotAxis = f.add_subplot(111)

    cmap1 = col.ListedColormap(color, 'temp')
    cm.register_cmap(cmap=cmap1)

    if zoom != 1:
        mask = ni.interpolation.zoom(mask, zoom, order=0)

    mask2 = mask.astype(np.float32)
    mask2[np.invert(np.isnan(mask2))] = 1.
    mask2[np.isnan(mask2)] = 0.

    struc = ni.generate_binary_structure(2, 2)
    if borderWidth:
        border = mask2 - ni.binary_erosion(mask2, struc, iterations=borderWidth).astype(np.float32)
    else:
        border = mask2 - ni.binary_erosion(mask2, struc).astype(np.float32)

    if closingIteration:
        border = ni.binary_closing(border, iterations=closingIteration).astype(np.float32)

    border[border == 0] = np.nan

    currfig = plotAxis.imshow(border, cmap='temp', interpolation='nearest')

    return currfig


def plot_mask_borders(mask, plotAxis=None, color='#ff0000', zoom=1, borderWidth=2, closingIteration=None, **kwargs):
    """
    plot mask (ROI) borders by using pyplot.contour function. all the 0s and Nans in the input mask will be considered
    as background, and non-zero, non-nan pixel will be considered in ROI.
    """
    if not plotAxis:
        f = plt.figure()
        plotAxis = f.add_subplot(111)

    plotingMask = np.ones(mask.shape, dtype=np.uint8)

    plotingMask[np.logical_or(np.isnan(mask), mask == 0)] = 0

    if zoom != 1:
        plotingMask = cv2.resize(plotingMask.astype(np.float),
                                 dsize=(int(plotingMask.shape[1] * zoom), int(plotingMask.shape[0] * zoom)))
        plotingMask[plotingMask < 0.5] = 0
        plotingMask[plotingMask >= 0.5] = 1
        plotingMask = plotingMask.astype(np.uint8)

    if closingIteration is not None:
        plotingMask = ni.binary_closing(plotingMask, iterations=closingIteration).astype(np.uint8)

    plotingMask = ni.binary_erosion(plotingMask, iterations=borderWidth)

    currfig = plotAxis.contour(plotingMask, levels=[0.5], colors=color, linewidths=borderWidth, **kwargs)

    # put y axis in decreasing order
    y_lim = list(plotAxis.get_ylim())
    y_lim.sort()
    plotAxis.set_ylim(y_lim[::-1])

    plotAxis.set_aspect('equal')

    return currfig


def grid_axis(rowNum, columnNum, totalPlotNum, **kwarg):
    """
    return figure handles and axis handels for multiple subplots and figures
    """

    figureNum = totalPlotNum // (rowNum * columnNum) + 1

    figureHandles = []

    for i in range(figureNum):
        f = plt.figure(**kwarg)
        figureHandles.append(f)

    axisHandles = []
    for i in range(totalPlotNum):
        currFig = figureHandles[i // (rowNum * columnNum)]
        currIndex = i % (rowNum * columnNum)
        currAxis = currFig.add_subplot(rowNum, columnNum, currIndex + 1)
        axisHandles.append(currAxis)

    return figureHandles, axisHandles


def tile_axis(f, rowNum, columnNum, topDownMargin=0.05, leftRightMargin=0.05, rowSpacing=0.05, columnSpacing=0.05):
    if 2 * topDownMargin + (
        (rowNum - 1) * rowSpacing) >= 1: raise ValueError, 'Top down margin or row spacing are too big!'
    if 2 * leftRightMargin + (
        (columnNum - 1) * columnSpacing) >= 1: raise ValueError, 'Left right margin or column spacing are too big!'

    height = (1 - (2 * topDownMargin) - (rowNum - 1) * rowSpacing) / rowNum
    width = (1 - (2 * leftRightMargin) - (columnNum - 1) * columnSpacing) / columnNum

    xStarts = np.arange(leftRightMargin, 1 - leftRightMargin, (width + columnSpacing))
    yStarts = np.arange(topDownMargin, 1 - topDownMargin, (height + rowSpacing))[::-1]

    axisList = [[f.add_axes([xStart, yStart, width, height]) for xStart in xStarts] for yStart in yStarts]

    return axisList


def save_figure_without_borders(f,
                                savePath,
                                removeSuperTitle=True,
                                **kwargs):
    """
    remove borders of a figure
    """
    f.gca().get_xaxis().set_visible(False)
    f.gca().get_yaxis().set_visible(False)
    f.gca().set_title('')
    if removeSuperTitle:
        f.suptitle('')
    f.savefig(savePath, pad_inches=0, bbox_inches='tight', **kwargs)


def merge_normalized_images(imgList, isFilter=True, sigma=50, mergeMethod='mean', dtype=np.float32):
    """
    merge images in a list in to one, for each image, local intensity variability will be removed by subtraction of
    gaussian filtered image. Then all images will be collapsed by the mergeMethod in to single image
    """

    imgList2 = []

    for currImg in imgList:
        imgList2.append(ia.array_nor(currImg.astype(dtype)))

    if mergeMethod == 'mean':
        mergedImg = np.mean(np.array(imgList2), axis=0)
    elif mergeMethod == 'min':
        mergedImg = np.min(np.array(imgList2), axis=0)
    elif mergeMethod == 'max':
        mergedImg = np.max(np.array(imgList2), axis=0)
    elif mergeMethod == 'median':
        mergedImg = np.median(np.array(imgList2), axis=0)

    if isFilter:
        mergedImgf = ni.filters.gaussian_filter(mergedImg.astype(np.float), sigma=sigma)
        return ia.array_nor(mergedImg - mergedImgf).astype(dtype)
    else:
        return ia.array_nor(mergedImg).astype(dtype)


# def hue2RGB(hue):
#     """
#     get the RGB value as format as hex string from the decimal ratio of hue (from 0 to 1)
#     color model as described in:
#     https://en.wikipedia.org/wiki/Hue
#     """
#     if hue < 0: hue = 0
#     if hue > 1: hue = 1
#     color = colorsys.hsv_to_rgb(hue,1,1)
#     color = [int(x*255) for x in color]
#     return get_color_str(*color)
#
#
def hot_2_rgb(hot):
    """
    get the RGB value as format as hex string from the decimal ratio of hot colormap (from 0 to 1)
    """
    if hot < 0: hot = 0
    if hot > 1: hot = 1
    cmap_hot = plt.get_cmap('hot')
    color = cmap_hot(hot)[0:3];
    color = [int(x * 255) for x in color]
    return get_color_str(*color)


def value_2_rgb(value, cmap):
    """
    get the RGB value as format as hex string from the decimal ratio of a given colormap (from 0 to 1)
    """
    if value < 0: value = 0
    if value > 1: value = 1
    cmap = plt.get_cmap(cmap)
    color = cmap(value)[0:3];
    color = [int(x * 255) for x in color]
    return get_color_str(*color)


if __name__ == '__main__':
    plt.ioff()
    print 'for debug'
